# Copyright (C) 2018-2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from __future__ import annotations

from abc import ABCMeta, abstractmethod
from collections import deque
import dataclasses
import heapq
from typing import TYPE_CHECKING, Any, Dict, Optional, Set, TypeVar

from swh.model.model import Sha1Git

if TYPE_CHECKING:
    from swh.storage.interface import StorageInterface


@dataclasses.dataclass
class State:
    done: Set[Sha1Git] = dataclasses.field(default_factory=set)
    revs_to_visit: Any = dataclasses.field(default_factory=list)
    last_rev: Optional[Dict] = None
    num_revs: int = 0
    missing_revs: Set[Sha1Git] = dataclasses.field(default_factory=set)


_revs_walker_classes = {}


class _RevisionsWalkerMetaClass(ABCMeta):
    def __new__(cls, clsname, bases, attrs):
        newclass = super().__new__(cls, clsname, bases, attrs)
        if "rw_type" in attrs:
            _revs_walker_classes[attrs["rw_type"]] = newclass
        return newclass


TWalker = TypeVar("TWalker", bound="RevisionsWalker")


class RevisionsWalker(metaclass=_RevisionsWalkerMetaClass):
    """
    Abstract base class encapsulating the logic to walk across
    a revisions history starting from a given one.

    It defines an iterator returning the revisions according
    to a specific ordering implemented in derived classes.

    The iteration step performs the following operations:

        1) Check if the iteration is finished by calling method
           :meth:`is_finished` and raises :exc:`StopIteration` if it
           it is the case

        2) Get the next unseen revision by calling method
           :meth:`get_next_rev_id`

        3) Process parents of that revision by calling method
           :meth:`process_parent_revs` for the next iteration
           steps

        4) Check if the revision should be returned by calling
           method :meth:`should_return` and returns it if
           it is the case

    In order to easily instantiate a specific type of revisions
    walker, it is recommended to use the factory function
    :func:`get_revisions_walker`.

    Args:
        storage: instance of swh storage (either local or remote)
        rev_start: a revision identifier
        max_revs: maximum number of revisions to return
        state: previous state of that revisions walker
        ignore_displayname: return the original author/committer's full name even if
          it's masked by a displayname.
    """

    def __init__(
        self,
        storage: StorageInterface,
        rev_start: Sha1Git,
        max_revs: Optional[int] = None,
        state: Optional[State] = None,
        ignore_displayname: bool = False,
    ):
        self._revs: Dict[Sha1Git, Dict] = {}
        self._max_revs = max_revs
        self._state = state or State()
        self.storage = storage
        self.ignore_displayname = ignore_displayname
        self.process_rev(rev_start)

    @abstractmethod
    def process_rev(self, rev_id: Sha1Git) -> None:
        """
        Abstract method whose purpose is to process a newly visited
        revision during the walk.
        Derived classes must implement it according to the desired
        method to walk across the revisions history (for instance
        through a dfs on the revisions DAG).

        Args:
            rev_id: the newly visited revision identifier
        """
        pass

    @abstractmethod
    def get_next_rev_id(self) -> Sha1Git:
        """
        Abstract method whose purpose is to return the next revision
        during the iteration.
        Derived classes must implement it according to the desired
        method to walk across the revisions history.
        """
        pass

    def process_parent_revs(self, rev: Dict) -> None:
        """
        Process the parents of a revision when it is iterated.
        The default implementation simply calls :meth:`process_rev`
        for each parent revision in the order they are declared.

        Args:
            rev (dict): A dict describing a revision as returned by
                :meth:`swh.storage.interface.StorageInterface.revision_get`
        """
        for parent_id in rev["parents"]:
            self.process_rev(parent_id)

    def should_return(self, rev: Dict) -> bool:
        """
        Filter out a revision to return if needed.
        Default implementation returns all iterated revisions.

        Args:
            rev (dict): A dict describing a revision as returned by
                :meth:`swh.storage.interface.StorageInterface.revision_get`

        Returns:
            bool: Whether to return the revision in the iteration
        """
        return True

    def is_finished(self) -> bool:
        """
        Determine if the iteration is finished.
        This method is called at the beginning of each iteration loop.

        Returns:
            bool: Whether the iteration is finished
        """
        if self._max_revs is not None and self._state.num_revs >= self._max_revs:
            return True
        if not self._state.revs_to_visit:
            return True
        return False

    def _get_rev(self, rev_id: Sha1Git) -> Optional[Dict]:
        rev = self._revs.get(rev_id)
        if rev is None:
            rev_obj = self.storage.revision_get(
                [rev_id], ignore_displayname=self.ignore_displayname
            )[0]
            if rev_obj is not None:
                self._revs[rev_obj.id] = rev_obj.to_dict()
                # prefetch parent revisions to speedup the revisions walk
                for parent in self.storage.revision_get(
                    list(set(rev_obj.parents) - self._revs.keys()),
                    ignore_displayname=self.ignore_displayname,
                ):
                    if parent is not None:
                        self._revs[parent.id] = parent.to_dict()
        return self._revs.get(rev_id)

    def missing_revisions(self) -> Set[Sha1Git]:
        """
        Return a set of revision identifiers whose associated data were
        found missing into the archive content while walking on the
        revisions graph.

        Returns:
            Set[bytes]: a set of revision identifiers
        """
        return self._state.missing_revs

    def is_history_truncated(self) -> bool:
        """
        Return if the revision history generated so far has been truncated
        of not. A revision history might end up truncated if some revision
        data were found missing into the archive content.

        Returns:
            bool: Whether the history got truncated or not
        """
        return len(self.missing_revisions()) > 0

    def export_state(self) -> State:
        """
        Export the internal state of that revision walker to a dict.
        Its purpose is to continue the iteration in a pagination context.

        Returns:
            The internal state of that revisions walker
        """
        return self._state

    def __next__(self) -> Dict:
        if self.is_finished():
            raise StopIteration
        while self._state.revs_to_visit:
            rev_id = self.get_next_rev_id()
            if rev_id in self._state.done:
                continue
            self._state.done.add(rev_id)
            rev = self._get_rev(rev_id)
            # revision data is missing, returned history will be truncated
            if rev is None:
                self._state.missing_revs.add(rev_id)
                continue
            self.process_parent_revs(rev)
            if self.should_return(rev):
                self._state.num_revs += 1
                self._state.last_rev = rev
                return rev
        raise StopIteration

    def __iter__(self: TWalker) -> TWalker:
        return self


class CommitterDateRevisionsWalker(RevisionsWalker):
    """
    Revisions walker that returns revisions in reverse chronological
    order according to committer date (same behaviour as ``git log``)
    """

    rw_type = "committer_date"

    def process_rev(self, rev_id: Sha1Git) -> None:
        """
        Add the revision to a priority queue according to the committer date.

        Args:
            rev_id (bytes): the newly visited revision identifier
        """
        if rev_id not in self._state.done:
            rev = self._get_rev(rev_id)
            if rev is not None:
                commit_time = (
                    rev["committer_date"]["timestamp"]["seconds"]
                    if rev["committer_date"]
                    # allows to avoid failure with a revision without commit date
                    # and iterate on such revision before its parents
                    else len(self._state.revs_to_visit)
                )
                heapq.heappush(self._state.revs_to_visit, (-commit_time, rev_id))
            else:
                self._state.missing_revs.add(rev_id)

    def get_next_rev_id(self) -> Sha1Git:
        """
        Return the smallest revision from the priority queue, i.e.
        the one with highest committer date.

        Returns:
            dict: A dict describing a revision as returned by
            :meth:`swh.storage.interface.StorageInterface.revision_get`
        """
        _, rev_id = heapq.heappop(self._state.revs_to_visit)
        return rev_id


class BFSRevisionsWalker(RevisionsWalker):
    """
    Revisions walker that returns revisions in the same order
    as when performing a breadth-first search on the revisions
    DAG.
    """

    rw_type = "bfs"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._state.revs_to_visit = deque(self._state.revs_to_visit)

    def process_rev(self, rev_id: Sha1Git) -> None:
        """
        Append the revision to a queue.

        Args:
            rev_id (bytes): the newly visited revision identifier
        """
        if rev_id not in self._state.done:
            self._state.revs_to_visit.append(rev_id)

    def get_next_rev_id(self) -> Sha1Git:
        """
        Return the next revision from the queue.

        Returns:
            dict: A dict describing a revision as returned by
            :meth:`swh.storage.interface.StorageInterface.revision_get`
        """
        return self._state.revs_to_visit.popleft()


class DFSPostRevisionsWalker(RevisionsWalker):
    """
    Revisions walker that returns revisions in the same order
    as when performing a depth-first search in post-order on the
    revisions DAG (i.e. after visiting a merge commit,
    the merged commit will be visited before the base it was
    merged on).
    """

    rw_type = "dfs_post"

    def process_rev(self, rev_id: Sha1Git) -> None:
        """
        Append the revision to a stack.

        Args:
            rev_id (bytes): the newly visited revision identifier
        """
        if rev_id not in self._state.done:
            self._state.revs_to_visit.append(rev_id)

    def get_next_rev_id(self) -> Sha1Git:
        """
        Return the next revision from the stack.

        Returns:
            dict: A dict describing a revision as returned by
            :meth:`swh.storage.interface.StorageInterface.revision_get`
        """
        return self._state.revs_to_visit.pop()


class DFSRevisionsWalker(DFSPostRevisionsWalker):
    """
    Revisions walker that returns revisions in the same order
    as when performing a depth-first search in pre-order on the
    revisions DAG (i.e. after visiting a merge commit,
    the base commit it was merged on will be visited before
    the merged commit).
    """

    rw_type = "dfs"

    def process_parent_revs(self, rev: Dict) -> None:
        """
        Process the parents of a revision when it is iterated in
        the reversed order they are declared.

        Args:
            rev (dict): A dict describing a revision as returned by
                :meth:`swh.storage.interface.StorageInterface.revision_get`
        """
        for parent_id in reversed(rev["parents"]):
            self.process_rev(parent_id)


class PathRevisionsWalker(CommitterDateRevisionsWalker):
    """
    Revisions walker that returns revisions where a specific
    path in the source tree has been modified, in other terms
    it allows to get the history for a specific file or directory.

    It has a behaviour similar to what ``git log`` offers by default,
    meaning the returned history is simplified in order to only
    show relevant revisions (see the `History Simplification
    <https://git-scm.com/docs/git-log#_history_simplification>`_
    section of the associated manual for more details).

    Please note that to avoid walking the entire history, the iteration
    will stop once a revision where the path has been added is found.

    .. warning:: Due to client-side implementation, performances
        are not optimal when the total numbers of revisions to walk
        is large. This should only be used when the total number of
        revisions does not exceed a couple of thousands.

    Args:
        storage (swh.storage.interface.StorageInterface): instance of swh storage
            (either local or remote)
        rev_start (bytes): a revision identifier
        path (str): the path in the source tree to retrieve the history
        max_revs (Optional[int]): maximum number of revisions to return
        state (Optional[dict]): previous state of that revisions walker
    """

    rw_type = "path"

    def __init__(self, storage, rev_start, path, **kwargs):
        super().__init__(storage, rev_start, **kwargs)
        paths = path.strip("/").split("/")
        self._path = list(map(lambda p: p.encode("utf-8"), paths))
        self._rev_dir_path = {}

    def _get_path_id(self, rev_id):
        """
        Return the path checksum identifier in the source tree of the
        provided revision. If the path corresponds to a directory, the
        value computed by :meth:`swh.model.Directory.compute_hash`
        will be returned. If the path corresponds to a file, its sha1
        checksum will be returned.

        Args:
            rev_id (bytes): a revision identifier

        Returns:
            bytes: the path identifier
        """

        rev = self._get_rev(rev_id)

        rev_dir_id = rev["directory"]

        if rev_dir_id not in self._rev_dir_path:
            try:
                dir_info = self.storage.directory_entry_get_by_path(
                    rev_dir_id, self._path
                )
                self._rev_dir_path[rev_dir_id] = dir_info["target"]
            except Exception:
                self._rev_dir_path[rev_dir_id] = None

        return self._rev_dir_path[rev_dir_id]

    def is_finished(self):
        """
        Check if the revisions iteration is finished.
        This checks for the specified path's existence in the last
        returned revision's parents' source trees.
        If not, the iteration is considered finished.

        Returns:
            bool: Whether to return the revision in the iteration
        """
        if self._path and self._last_rev:
            last_rev_parents = self._last_rev["parents"]
            last_rev_parents_path_ids = [
                self._get_path_id(p_rev) for p_rev in last_rev_parents
            ]
            no_path = all([path_id is None for path_id in last_rev_parents_path_ids])
            if no_path:
                return True
        return super().is_finished()

    def process_parent_revs(self, rev):
        """
        Process parents when a new revision is iterated.
        It enables to get a simplified revisions history in the same
        manner as ``git log``. When a revision has multiple parents,
        the following process is applied. If the revision was a merge,
        and has the same path identifier to one parent, follow only that
        parent (even if there are several parents with the same path
        identifier, follow only one of them.) Otherwise, follow all parents.

        Args:
            rev (dict): A dict describing a revision as returned by
                :meth:`swh.storage.interface.StorageInterface.revision_get`
        """
        rev_path_id = self._get_path_id(rev["id"])

        if rev_path_id:
            if len(rev["parents"]) == 1:
                self.process_rev(rev["parents"][0])
            else:
                parent_rev_path_ids = [
                    self._get_path_id(p_rev) for p_rev in rev["parents"]
                ]
                different_trees = all(
                    [path_id != rev_path_id for path_id in parent_rev_path_ids]
                )
                for i, p_rev in enumerate(rev["parents"]):
                    if different_trees or parent_rev_path_ids[i] == rev_path_id:
                        self.process_rev(p_rev)
                        if not different_trees:
                            break
        else:
            super().process_parent_revs(rev)

    def should_return(self, rev):
        """
        Check if a revision should be returned when iterating.
        It verifies that the specified path has been modified
        by the revision but also that all parents have a path
        identifier different from the revision one in order
        to get a simplified history.

        Args:
            rev (dict): A dict describing a revision as returned by
                :meth:`swh.storage.interface.StorageInterface.revision_get`

        Returns:
            bool: Whether to return the revision in the iteration
        """
        rev_path_id = self._get_path_id(rev["id"])

        if not rev["parents"]:
            return rev_path_id is not None

        parent_rev_path_ids = [self._get_path_id(p_rev) for p_rev in rev["parents"]]
        different_trees = all(
            [path_id != rev_path_id for path_id in parent_rev_path_ids]
        )

        if rev_path_id != parent_rev_path_ids[0] and different_trees:
            return True

        return False


def get_revisions_walker(rev_walker_type, *args, **kwargs):
    """
    Instantiate a revisions walker of a given type.

    The following code snippet demonstrates how to use a revisions
    walker for processing a whole revisions history::

        from swh.storage import get_storage

        storage = get_storage(...)

        revs_walker = get_revisions_walker('committer_date', storage, rev_id)
        for rev in revs_walker:
            # process revision rev

    It is also possible to walk a revisions history in a paginated
    way as illustrated below::

        def get_revs_history_page(rw_type, storage, rev_id, page_num,
                                  page_size, rw_state):
            max_revs = (page_num + 1) * page_size
            revs_walker = get_revisions_walker(rw_type, storage, rev_id,
                                               max_revs=max_revs,
                                               state=rw_state)
            revs = list(revs_walker)
            rw_state = revs_walker.export_state()
            return revs

        rev_start = ...
        per_page = 50
        rw_state = {}

        for page in range(0, 10):
            revs_page = get_revs_history_page('dfs', storage, rev_start, page,
                                              per_page, rw_state)
            # process revisions page


    Args:
        rev_walker_type (str): the type of revisions walker to return,
            possible values are: *committer_date*, *dfs*, *dfs_post*,
            *bfs* and *path*
        args (list): position arguments to pass to the revisions walker
            constructor
        kwargs (dict): keyword arguments to pass to the revisions walker
            constructor

    """
    if rev_walker_type not in _revs_walker_classes:
        raise Exception('No revisions walker found for type "%s"' % rev_walker_type)
    revs_walker_class = _revs_walker_classes[rev_walker_type]
    return revs_walker_class(*args, **kwargs)



# Copyright (C) 2020-2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""Decoder and encoders for swh-model objects."""

from typing import Any, Callable, Dict, List, Tuple
import uuid

from swh.model import model, swhids
from swh.storage import interface
from swh.storage.proxies.blocking.db import BlockingState, BlockingStatus
from swh.storage.proxies.masking.db import MaskedState, MaskedStatus


def _encode_model_object(obj):
    d = obj.to_dict()
    d["__type__"] = type(obj).__name__
    return d


def _encode_enum(obj):
    return {
        "value": obj.value,
        "__type__": type(obj).__name__,
    }


def _encode_origin_visit_with_statuses(
    ovws: interface.OriginVisitWithStatuses,
) -> Dict[str, Any]:
    return {
        "visit": ovws.visit.to_dict(),
        "statuses": [status.to_dict() for status in ovws.statuses],
    }


def _encode_object_reference(object_reference: interface.ObjectReference) -> List[str]:
    return [str(object_reference.source), str(object_reference.target)]


def _encode_snapshot_branch_by_name_response(
    branch_by_name_response: interface.SnapshotBranchByNameResponse,
) -> Dict[str, Any]:
    target = branch_by_name_response.target
    return {
        "branch_found": branch_by_name_response.branch_found,
        "target": target.to_dict() if target else None,
        "aliases_followed": branch_by_name_response.aliases_followed,
    }


def _encode_masked_status(masked_status: MaskedStatus):
    return {
        "state": masked_status.state.name,
        "request": str(masked_status.request),
    }


def _decode_masked_status(d: Dict[str, Any]):
    return MaskedStatus(state=MaskedState[d["state"]], request=uuid.UUID(d["request"]))


def _encode_blocking_status(blocking_status: BlockingStatus):
    return {
        "state": blocking_status.state.name,
        "request": str(blocking_status.request),
    }


def _decode_blocking_status(d: Dict[str, Any]):
    return BlockingStatus(
        state=BlockingState[d["state"]], request=uuid.UUID(d["request"])
    )


def _decode_origin_visit_with_statuses(
    ovws: Dict[str, Any],
) -> interface.OriginVisitWithStatuses:
    return interface.OriginVisitWithStatuses(
        visit=model.OriginVisit(**ovws["visit"]),
        statuses=[model.OriginVisitStatus(**status) for status in ovws["statuses"]],
    )


def _decode_model_enum(d):
    return getattr(model, d.pop("__type__"))(d["value"])


def _decode_swhids_enum(d):
    return getattr(swhids, d.pop("__type__"))(d["value"])


def _decode_storage_enum(d):
    return getattr(interface, d.pop("__type__"))(d["value"])


def _decode_object_reference(d):
    return interface.ObjectReference(
        source=swhids.ExtendedSWHID.from_string(d[0]),
        target=swhids.ExtendedSWHID.from_string(d[1]),
    )


def _decode_snapshot_branch_by_name_response(d):
    target = model.SnapshotBranch.from_dict(d["target"]) if d["target"] else None
    return interface.SnapshotBranchByNameResponse(
        branch_found=d["branch_found"],
        target=target,
        aliases_followed=d["aliases_followed"],
    )


ENCODERS: List[Tuple[type, str, Callable]] = [
    (model.BaseModel, "model", _encode_model_object),
    (swhids.CoreSWHID, "core_swhid", str),
    (swhids.ExtendedSWHID, "extended_swhid", str),
    (swhids.QualifiedSWHID, "qualified_swhid", str),
    # TODO: serialize this as "swhids_enum" when all peers support it in their DECODERS:
    (swhids.ObjectType, "identifiers_enum", _encode_enum),
    (model.MetadataAuthorityType, "model_enum", _encode_enum),
    (interface.ListOrder, "storage_enum", _encode_enum),
    (
        interface.OriginVisitWithStatuses,
        "origin_visit_with_statuses",
        _encode_origin_visit_with_statuses,
    ),
    (
        interface.ObjectReference,
        "object_reference",
        _encode_object_reference,
    ),
    (
        interface.SnapshotBranchByNameResponse,
        "branch_by_name_response",
        _encode_snapshot_branch_by_name_response,
    ),
    (MaskedStatus, "masked_status", _encode_masked_status),
    (BlockingStatus, "blocking_status", _encode_blocking_status),
]


DECODERS: Dict[str, Callable] = {
    "core_swhid": swhids.CoreSWHID.from_string,
    "extended_swhid": swhids.ExtendedSWHID.from_string,
    "qualified_swhid": swhids.QualifiedSWHID.from_string,
    "model": lambda d: getattr(model, d.pop("__type__")).from_dict(d),
    "identifiers_enum": _decode_swhids_enum,
    "swhids_enum": _decode_swhids_enum,
    "model_enum": _decode_model_enum,
    "storage_enum": _decode_storage_enum,
    "origin_visit_with_statuses": _decode_origin_visit_with_statuses,
    "object_reference": _decode_object_reference,
    "branch_by_name_response": _decode_snapshot_branch_by_name_response,
    "masked_status": _decode_masked_status,
    "blocking_status": _decode_blocking_status,
}



# Copyright (C) 2024-2025 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from contextlib import contextmanager
import functools
import inspect
import itertools
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    TypeVar,
    Union,
)
import warnings

import attr
import psycopg_pool

from swh.core.utils import grouper
from swh.model.hashutil import MultiHash
from swh.model.model import (
    ExtID,
    Origin,
    Person,
    RawExtrinsicMetadata,
    Release,
    Revision,
    Sha1Git,
)
from swh.model.swhids import ExtendedObjectType, ExtendedSWHID
from swh.storage import get_storage
from swh.storage.exc import MaskedObjectException
from swh.storage.interface import HashDict, PagedResult, Sha1, StorageInterface, TResult
from swh.storage.metrics import DifferentialTimer
from swh.storage.proxies.masking.db import MaskedStatus

from .db import MaskingQuery

BATCH_SIZE = 1024
MASKING_OVERHEAD_METRIC = "swh_storage_masking_overhead_seconds"


def get_datastore(cls, db=None, masking_db=None, **kwargs):
    assert cls in ("postgresql", "masking")
    from .db import MaskingAdmin

    if db is None:
        db = masking_db
    return MaskingAdmin.connect(db)


def masking_overhead_timer(method_name: str) -> DifferentialTimer:
    """Return a properly setup DifferentialTimer for ``method_name`` of the storage"""
    return DifferentialTimer(MASKING_OVERHEAD_METRIC, tags={"endpoint": method_name})


class MaskingProxyStorage:
    """Masking storage proxy

    This proxy can return modified objects or stop them from being retrieved at
    all.

    It uses a specific PostgreSQL database (which for now is colocated with the
    swh.storage PostgreSQL database), the access to which is implemented in the
    :mod:`..masking.db` submodule.

    Sample configuration

    .. code-block: yaml

        storage:
          cls: masking
          db: 'dbname=swh-storage'
          max_pool_conns: 10
          storage:
          - cls: remote
            url: http://storage.internal.staging.swh.network:5002/

    """

    def __init__(
        self,
        storage: Union[Dict, StorageInterface],
        db: Optional[str] = None,
        masking_db: Optional[str] = None,
        min_pool_conns: int = 1,
        max_pool_conns: int = 5,
    ):
        if db is None:
            assert masking_db is not None
            warnings.warn(
                "'masking_db' field in the masking storage configuration "
                "was renamed 'db' field",
                DeprecationWarning,
            )
            db = masking_db
        self.storage: StorageInterface = (
            get_storage(**storage) if isinstance(storage, dict) else storage
        )

        self._masking_pool = psycopg_pool.ConnectionPool(
            db,
            min_size=min_pool_conns,
            max_size=max_pool_conns,
            open=True,
        )

        # Generate the method dictionaries once per instantiation, instead of
        # doing it on every (first) __getattr__ call.
        self._gen_method_dicts()

    @contextmanager
    def _masking_query(self) -> Iterator[MaskingQuery]:
        ret = None
        try:
            ret = MaskingQuery.from_pool(self._masking_pool)
            yield ret
        finally:
            if ret:
                ret.put_conn()

    @staticmethod
    def _get_swhids_in_result(method_name: str, result: Any) -> List[ExtendedSWHID]:
        if result is None:
            raise TypeError(f"Filtering of Nones missing in {method_name}")

        result_type = getattr(result, "object_type", None)
        if result_type == RawExtrinsicMetadata.object_type:
            # Raw Extrinsic Metadata have a swhid, but we also mask them if the target is masked
            return [result.swhid(), result.target]

        if hasattr(result, "swhid"):
            swhid = result.swhid()
            if hasattr(swhid, "to_extended"):
                swhid = swhid.to_extended()
            assert isinstance(
                swhid, ExtendedSWHID
            ), f"{method_name} returned object with unexpected swhid() method return type"
            return [swhid]

        if result_type:
            if result_type == ExtID.object_type:
                return [result.target.to_extended()]
            if result_type.value.startswith("origin_visit"):
                return [Origin(url=result.origin).swhid()]

        if (
            object_type := method_name.removesuffix("_get_random").removesuffix(
                "_get_id_partition"
            )
        ) != method_name:
            # Returns bare sha1_gits
            assert isinstance(result, bytes), f"{method_name} returned unexpected type"
            return [
                ExtendedSWHID(
                    object_type=ExtendedObjectType[object_type.upper()],
                    object_id=result,
                )
            ]
        elif method_name == "revision_log":
            # returns dicts of revisions
            assert (
                isinstance(result, dict) and "id" in result
            ), f"{method_name} returned unexpected type"
            return [
                ExtendedSWHID(
                    object_type=ExtendedObjectType.REVISION, object_id=result["id"]
                )
            ]
        elif method_name == "revision_shortlog":
            # Returns tuples (revision.id, revision.parents)
            assert isinstance(
                result[0], bytes
            ), f"{method_name} returned unexpected type"
            return [
                ExtendedSWHID(
                    object_type=ExtendedObjectType.REVISION, object_id=result[0]
                )
            ]

        elif method_name == "origin_get_by_sha1":
            # Returns origin dicts, because why not
            assert (
                isinstance(result, dict) and "url" in result
            ), f"{method_name} returned unexpected type"
            return [Origin(url=result["url"]).swhid()]
        elif method_name == "origin_visit_get_with_statuses":
            # Returns an OriginVisitWithStatuses
            return [Origin(url=result.visit.origin).swhid()]
        elif method_name == "snapshot_get":
            # Returns a snapshot dict
            return [
                ExtendedSWHID(
                    object_type=ExtendedObjectType.SNAPSHOT, object_id=result["id"]
                )
            ]

        raise ValueError(f"Cannot get swhid for result of method {method_name}")

    def _masked_result(
        self, method_name: str, result: Any
    ) -> Optional[Dict[ExtendedSWHID, List[MaskedStatus]]]:
        """Find the SWHIDs of the ``result`` object, and check if any of them is
        masked, returning the associated masking information."""
        with self._masking_query() as q:
            return q.swhids_are_masked(self._get_swhids_in_result(method_name, result))

    def _raise_if_masked_result(self, method_name: str, result: Any) -> None:
        """Raise a :exc:`MaskedObjectException` if ``result`` is masked."""
        masked = self._masked_result(method_name, result)
        if masked:
            raise MaskedObjectException(masked)

    def _raise_if_masked_swhids(self, swhids: List[ExtendedSWHID]) -> None:
        """Raise a :exc:`MaskedObjectException` if any SWHID is masked."""
        with self._masking_query() as q:
            masked = q.swhids_are_masked(swhids)

        if masked:
            raise MaskedObjectException(masked)

    def __getattr__(self, key):
        method = None
        if key in self._methods_by_name:
            method = self._methods_by_name[key](key)
        else:
            suffix = key.rsplit("_", 1)[-1]

            if suffix in self._methods_by_suffix:
                method = self._methods_by_suffix[suffix](key)

        if method:
            # Avoid going through __getattr__ again next time
            setattr(self, key, method)
            return method

        # Raise a NotImplementedError to make sure we don't forget to add
        # masking to any new storage functions
        raise NotImplementedError(key)

    def _gen_method_dicts(self):
        """Generate the :attr:`_methods_by_name` and :attr:`_methods_by_suffix`
        used by :meth:`__getattr__`"""
        _passthrough = functools.partial(getattr, self.storage)

        self._methods_by_name = {
            # Returns a single object
            "snapshot_get": self._getter_optional,
            "origin_visit_find_by_date": self._getter_optional,
            "origin_visit_get_by": self._getter_optional,
            "origin_visit_status_get_latest": self._getter_optional,
            "origin_visit_get_latest": self._getter_optional,
            # Returns a PagedResult
            "origin_list": self._getter_pagedresult,
            "origin_visit_get": self._getter_pagedresult,
            "origin_search": self._getter_pagedresult,
            "raw_extrinsic_metadata_get": self._getter_pagedresult,
            "origin_visit_get_with_statuses": self._getter_pagedresult,
            "origin_visit_status_get": self._getter_pagedresult,
            # Returns a list of (optional) objects
            "origin_get_by_sha1": self._getter_list,
            "content_find": self._getter_list,
            "skipped_content_find": self._getter_list,
            "revision_shortlog": self._getter_list,
            "extid_get_from_target": self._getter_list,
            "raw_extrinsic_metadata_get_by_ids": self._getter_list,
            # Filter arguments
            "directory_entry_get_by_path": self._getter_filtering_arguments,
            "directory_get_entries": self._getter_filtering_arguments,
            "directory_get_raw_manifest": self._getter_filtering_arguments,
            "directory_ls": self._getter_filtering_arguments,
            "raw_extrinsic_metadata_get_authorities": self._getter_filtering_arguments,
            "snapshot_branch_get_by_name": self._getter_filtering_arguments,
            "snapshot_count_branches": self._getter_filtering_arguments,
            "snapshot_get_branches": self._getter_filtering_arguments,
            "origin_snapshot_get_all": self._getter_filtering_arguments,
            # Content functions that don't match common getter or adder suffixes
            "content_add_metadata": _passthrough,
            "content_missing_per_sha1": _passthrough,
            "content_missing_per_sha1_git": _passthrough,
            "content_update": _passthrough,
            # These objects aren't maskable
            "extid_get_from_extid": _passthrough,
            "object_find_by_sha1_git": _passthrough,
            "object_find_recent_references": _passthrough,
            "metadata_authority_get": _passthrough,
            "metadata_fetcher_get": _passthrough,
            # Utility methods
            "check_config": _passthrough,
            "clear_buffers": _passthrough,
            "flush": _passthrough,
            "origin_count": _passthrough,
            "refresh_stat_counters": _passthrough,
            "stat_counters": _passthrough,
            # For tests
            "journal_writer": _passthrough,
        }

        self._methods_by_suffix = {
            # These methods will never need do any masking
            "add": _passthrough,
            "missing": _passthrough,
            # Partitions return PagedResults
            "partition": self._getter_pagedresult,
            # Getters return lists of optional objects
            "get": self._getter_list,
            "random": self._getter_random,
        }

    def content_get_data(self, content: Union[HashDict, Sha1]) -> Optional[bytes]:
        with masking_overhead_timer("content_get_data") as t:
            with t.inner():
                ret = self.storage.content_get_data(content)

            if ret is None:
                return None

            if isinstance(content, dict) and "sha1_git" in content:
                self._raise_if_masked_swhids(
                    [
                        ExtendedSWHID(
                            object_type=ExtendedObjectType.CONTENT,
                            object_id=content["sha1_git"],
                        )
                    ]
                )

            else:
                # We did not get the SWHID of the object as argument, so we need to
                # hash the resulting content to check if its SWHID was masked.
                self._raise_if_masked_swhids(
                    [
                        ExtendedSWHID(
                            object_type=ExtendedObjectType.CONTENT,
                            object_id=MultiHash.from_data(ret, ["sha1_git"]).digest()[
                                "sha1_git"
                            ],
                        )
                    ]
                )

            return ret

    def _get_swhids_in_args(
        self, method_name: str, parsed_args: Dict[str, Any]
    ) -> List[ExtendedSWHID]:
        """Extract SWHIDs from the parsed arguments of ``method_name``.

        Arguments:
          method_name: name of the called method
          parsed_args: arguments of the method parsed with :func:`inspect.getcallargs`
        """

        if method_name in ("directory_entry_get_by_path", "directory_ls"):
            return [
                ExtendedSWHID(
                    object_type=ExtendedObjectType.DIRECTORY,
                    object_id=parsed_args["directory"],
                )
            ]
        elif method_name == "directory_get_entries":
            return [
                ExtendedSWHID(
                    object_type=ExtendedObjectType.DIRECTORY,
                    object_id=parsed_args["directory_id"],
                )
            ]
        elif method_name.startswith("snapshot_"):
            return [
                ExtendedSWHID(
                    object_type=ExtendedObjectType.SNAPSHOT,
                    object_id=parsed_args["snapshot_id"],
                )
            ]
        elif method_name == "raw_extrinsic_metadata_get_authorities":
            return [parsed_args["target"]]
        elif method_name == "directory_get_raw_manifest":
            return [
                ExtendedSWHID(
                    object_type=ExtendedObjectType.DIRECTORY,
                    object_id=object_id,
                )
                for object_id in parsed_args["directory_ids"]
            ]
        elif method_name == "origin_snapshot_get_all":
            return [Origin(url=parsed_args["origin_url"]).swhid()]
        else:
            raise ValueError(f"Cannot get swhid for arguments of method {method_name}")

    def _getter_filtering_arguments(self, method_name: str):
        """Handles methods that should filter on their argument, instead of the
        returned value. If the underlying storage returns :const:`None`, return
        it, else, raise :exc:`MaskedObjectException` if the requested object is
        masked"""

        @functools.wraps(getattr(self.storage, method_name))
        def newf(*args, **kwargs):
            with masking_overhead_timer(method_name) as t:
                method = getattr(self.storage, method_name)

                with t.inner():
                    result = method(*args, **kwargs)

                if result is None:
                    return None

                signature = inspect.signature(getattr(StorageInterface, method_name))
                bound_args = signature.bind(self, *args, **kwargs)
                self._raise_if_masked_swhids(
                    self._get_swhids_in_args(method_name, bound_args.arguments)
                )

                return result

        return newf

    RANDOM_ATTEMPTS = 5

    def _getter_random(self, method_name: str):
        """Handles methods returning a random object. Try
        :const:`RANDOM_ATTEMPTS` times for a non-masked object, and return it,
        else return :const:`None`."""

        @functools.wraps(getattr(self.storage, method_name))
        def newf(*args, **kwargs):
            with masking_overhead_timer(method_name) as t:
                method = getattr(self.storage, method_name)
                for _ in range(self.RANDOM_ATTEMPTS):
                    with t.inner():
                        result = method(*args, **kwargs)

                    if result is None:
                        return None

                    if not self._masked_result(method_name, result):
                        return result

        return newf

    def _getter_optional(self, method_name: str):
        """Handles methods returning an optional object: if the return value is
        :const:`None`, return it, else, raise a :exc:`MaskedObjectException` if
        the return value should be masked."""

        @functools.wraps(getattr(self.storage, method_name))
        def newf(*args, **kwargs):
            with masking_overhead_timer(method_name) as t:
                method = getattr(self.storage, method_name)
                with t.inner():
                    result = method(*args, **kwargs)
                if result is None:
                    return None

                self._raise_if_masked_result(method_name, result)

                return result

        return newf

    def _raise_if_masked_result_in_list(
        self, method_name: str, results: Iterable[TResult]
    ) -> List[TResult]:
        """Raise a :exc:`MaskedObjectException` if any non-:const:`None` object
        in ``results`` is masked."""
        result_swhids = set()
        results = list(results)
        for result in results:
            if result is not None:
                result_swhids.update(self._get_swhids_in_result(method_name, result))

        if result_swhids:
            self._raise_if_masked_swhids(list(result_swhids))
        return results

    def _getter_list(
        self,
        method_name: str,
    ):
        """Handle methods returning a list (or a generator) of optional objects,
        raising :exc:`MaskedObjectException` for all the masked objects in the
        batch."""

        @functools.wraps(getattr(self.storage, method_name))
        def newf(*args, **kwargs):
            with masking_overhead_timer(method_name) as t:
                method = getattr(self.storage, method_name)

                with t.inner():
                    results = list(method(*args, **kwargs))

                self._raise_if_masked_result_in_list(method_name, results)
                return results

        return newf

    def _getter_pagedresult(self, method_name: str) -> Callable[..., PagedResult]:
        """Handle methods returning a :cls:`PagedResult`, raising
        :exc:`MaskedObjectException` if some objects in the returned page are
        masked."""

        @functools.wraps(getattr(self.storage, method_name))
        def newf(*args, **kwargs) -> PagedResult:
            with masking_overhead_timer(method_name) as t:
                method = getattr(self.storage, method_name)
                with t.inner():
                    results = method(*args, **kwargs)

                self._raise_if_masked_result_in_list(method_name, results.results)

                return results

        return newf

    # Patching proxy feature set

    TRevision = TypeVar("TRevision", Revision, Optional[Revision])

    def _apply_revision_display_names(
        self, revisions: List[TRevision]
    ) -> List[TRevision]:
        emails = set()
        for rev in revisions:
            if (
                rev is not None
                and rev.author is not None
                and rev.author.email  # ignore None or empty email addresses
            ):
                emails.add(rev.author.email)
            if (
                rev is not None
                and rev.committer is not None
                and rev.committer.email  # ignore None or empty email addresses
            ):
                emails.add(rev.committer.email)

        with self._masking_query() as q:
            display_names = q.display_name(list(emails))

        # Short path for the common case
        if not display_names:
            return revisions

        persons: Dict[Optional[bytes], Person] = {
            email: Person.from_fullname(display_name)
            for (email, display_name) in display_names.items()
        }

        return [
            (
                None
                if revision is None
                else attr.evolve(
                    revision,
                    author=(
                        revision.author
                        if revision.author is None
                        else persons.get(revision.author.email, revision.author)
                    ),
                    committer=(
                        revision.committer
                        if revision.committer is None
                        else persons.get(revision.committer.email, revision.committer)
                    ),
                )
            )
            for revision in revisions
        ]

    TRelease = TypeVar("TRelease", Release, Optional[Release])

    def _apply_release_display_names(self, releases: List[TRelease]) -> List[TRelease]:
        emails = set()
        for rel in releases:
            if (
                rel is not None
                and rel.author is not None
                and rel.author.email  # ignore None or empty email addresses
            ):
                emails.add(rel.author.email)

        with self._masking_query() as q:
            display_names = q.display_name(list(emails))
        # Short path for the common case
        if not display_names:
            return releases

        persons: Dict[Optional[bytes], Person] = {
            email: Person.from_fullname(display_name)
            for (email, display_name) in display_names.items()
        }

        return [
            (
                None
                if release is None
                else attr.evolve(
                    release,
                    author=(
                        release.author
                        if release.author is None
                        else persons.get(release.author.email, release.author)
                    ),
                )
            )
            for release in releases
        ]

    def revision_get(
        self, revision_ids: List[Sha1Git], ignore_displayname: bool = False
    ) -> List[Optional[Revision]]:
        revisions = self.storage.revision_get(revision_ids)
        self._raise_if_masked_result_in_list("revision_get", revisions)
        return self._apply_revision_display_names(revisions)

    def revision_log(
        self,
        revisions: List[Sha1Git],
        ignore_displayname: bool = False,
        limit: Optional[int] = None,
    ) -> Iterable[Optional[Dict[str, Any]]]:
        revision_batches = grouper(
            self.storage.revision_log(revisions, limit=limit), BATCH_SIZE
        )
        yield from map(
            Revision.to_dict,
            itertools.chain.from_iterable(
                self._apply_revision_display_names(
                    self._raise_if_masked_result_in_list(
                        "revision_log", list(map(Revision.from_dict, revision_batch))
                    )
                )
                for revision_batch in revision_batches
            ),
        )

    def revision_get_partition(
        self,
        partition_id: int,
        nb_partitions: int,
        page_token: Optional[str] = None,
        limit: int = 1000,
    ) -> PagedResult[Revision]:
        page: PagedResult[Revision] = self.storage.revision_get_partition(
            partition_id, nb_partitions, page_token, limit
        )
        return PagedResult(
            results=self._apply_revision_display_names(
                self._raise_if_masked_result_in_list(
                    "revision_get_parition", page.results
                )
            ),
            next_page_token=page.next_page_token,
        )

    def release_get(
        self, releases: List[Sha1Git], ignore_displayname: bool = False
    ) -> List[Optional[Release]]:
        return self._apply_release_display_names(
            self._raise_if_masked_result_in_list(
                "release_get", self.storage.release_get(releases)
            )
        )

    def release_get_partition(
        self,
        partition_id: int,
        nb_partitions: int,
        page_token: Optional[str] = None,
        limit: int = 1000,
    ) -> PagedResult[Release]:
        page = self.storage.release_get_partition(
            partition_id, nb_partitions, page_token, limit
        )
        return PagedResult(
            results=self._apply_release_display_names(
                self._raise_if_masked_result_in_list(
                    "release_get_partition", page.results
                )
            ),
            next_page_token=page.next_page_token,
        )



# Copyright (C) 2019-2025 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import dataclasses
import datetime
import functools
import logging
import re
from typing import Any, Container, Dict, List, Optional, Tuple, cast

import attr
import pytest
import yaml

from swh.journal.client import EofBehavior, JournalClient
from swh.journal.serializers import kafka_to_value, key_to_kafka, value_to_kafka
from swh.journal.writer import JournalWriterInterface
from swh.model.hashutil import MultiHash, hash_to_bytes
from swh.model.model import (
    Directory,
    OriginVisit,
    OriginVisitStatus,
    Release,
    Revision,
    RevisionType,
    SkippedContent,
    Snapshot,
)
from swh.model.tests.swh_model_data import COMMITTERS, DATES, REVISIONS
from swh.model.tests.swh_model_data import TEST_OBJECTS as _TEST_OBJECTS
from swh.storage import get_storage
from swh.storage.cassandra.model import ContentRow, SkippedContentRow
from swh.storage.exc import StorageArgumentException
from swh.storage.in_memory import InMemoryStorage
from swh.storage.interface import StorageInterface
from swh.storage.replay import ModelObjectDeserializer, process_replay_objects
from swh.storage.tests.storage_data import StorageData

UTC = datetime.timezone.utc

TEST_OBJECTS = _TEST_OBJECTS.copy()
# Add a revision with metadata to check this later is dropped while being replayed
# TODO: replace this with a lower level struct (if still needed) since the
# metadata field is about to be removed from the Revision object model
TEST_OBJECTS[Revision.object_type] = list(_TEST_OBJECTS[Revision.object_type]) + [
    Revision(
        id=hash_to_bytes("51d9d94ab08d3f75512e3a9fd15132e0a7ca7928"),
        message=b"hello again",
        date=DATES[1],
        committer=COMMITTERS[1],
        author=COMMITTERS[0],
        committer_date=DATES[0],
        type=RevisionType.GIT,
        directory=b"\x03" * 20,
        synthetic=False,
        metadata={"something": "interesting"},
        parents=(REVISIONS[0].id,),
    ),
]
WRONG_ID_REG = re.compile(
    "Object has id [0-9a-f]{40}, but it should be [0-9a-f]{40}: .*"
)


def now():
    return datetime.datetime.now(datetime.UTC)


def nullify_ctime(obj: Any) -> Any:
    if isinstance(obj, (ContentRow, SkippedContentRow)):
        return dataclasses.replace(obj, ctime=None)
    else:
        return obj


@pytest.fixture()
def replayer_storage(kafka_prefix: str, kafka_server: str):
    journal_writer_config = {
        "cls": "kafka",
        "brokers": [kafka_server],
        "client_id": "kafka_writer",
        "prefix": kafka_prefix,
        "auto_flush": False,
    }
    storage_config: Dict[str, Any] = {
        "cls": "memory",
        "journal_writer": journal_writer_config,
    }
    return get_storage(**storage_config)


@pytest.fixture()
def replayer_client(kafka_prefix: str, kafka_consumer_group: str, kafka_server: str):
    deserializer = ModelObjectDeserializer()
    return JournalClient(
        brokers=kafka_server,
        group_id=kafka_consumer_group,
        prefix=kafka_prefix,
        on_eof=EofBehavior.STOP,
        value_deserializer=deserializer.convert,
    )


@pytest.fixture()
def replayer_storage_and_client(replayer_storage, replayer_client):
    # for bw compat
    return replayer_storage, replayer_client


@pytest.mark.parametrize(
    "tenacious",
    [pytest.param(True, id="tenacious"), pytest.param(False, id="non-tenacious")],
)
def test_storage_replayer(replayer_storage_and_client, caplog, tenacious):
    """Optimal replayer scenario.

    This:
    - writes objects to a source storage
    - replayer consumes objects from the topic and replays them
    - a destination storage is filled from this

    In the end, both storages should have the same content.
    """
    src, replayer = replayer_storage_and_client

    # Fill Kafka using a source storage
    nb_sent = 0
    for object_type, objects in TEST_OBJECTS.items():
        method = getattr(src, f"{object_type}_add")
        method(objects)
        nb_sent += len(objects)
    src.journal_writer.journal.flush()

    caplog.set_level(logging.ERROR, "swh.journal.replay")

    if tenacious:
        proxied_dst = get_storage("tenacious", storage={"cls": "memory"})
        dst = proxied_dst.storage
    else:
        proxied_dst = dst = get_storage(cls="memory")

    # Fill the destination storage from Kafka
    worker_fn = functools.partial(process_replay_objects, storage=proxied_dst)
    nb_inserted = replayer.process(worker_fn)
    assert nb_sent == nb_inserted

    assert isinstance(src, InMemoryStorage)  # needed to help mypy
    assert isinstance(dst, InMemoryStorage)
    check_replayed(src, dst)

    collision = 0
    for record in caplog.records:
        logtext = record.getMessage()
        if "Colliding contents:" in logtext:
            collision += 1

    assert collision == 0, "No collision should be detected"


def test_storage_replay_with_collision(
    replayer_storage_and_client: Tuple[StorageInterface, JournalClient], caplog
):
    """Another replayer scenario with collisions.

    This:
    - writes objects to the topic, including colliding contents
    - replayer consumes objects from the topic and replay them
    - This drops the colliding contents from the replay when detected

    """
    src, replayer = replayer_storage_and_client
    journal: JournalWriterInterface = src.journal_writer.journal  # type: ignore
    assert journal is not None

    # Fill Kafka using a source storage
    nb_sent = 0
    for object_type, objects in TEST_OBJECTS.items():
        method = getattr(src, f"{object_type}_add")
        method(objects)
        nb_sent += len(objects)
    journal.flush()

    # Create collision in input data
    # These should not be written in the destination
    producer = journal.producer  # type: ignore
    prefix = journal._prefix  # type: ignore
    DUPLICATE_CONTENTS = StorageData().colliding_contents["sha1"]
    for content in DUPLICATE_CONTENTS:
        topic = f"{prefix}.content"
        key = content.hashes()
        value = content.to_dict()
        value.pop("data", None)
        producer.produce(
            topic=topic,
            key=key_to_kafka(key),
            value=value_to_kafka(value),
        )
        nb_sent += 1

    producer.flush()

    caplog.set_level(logging.ERROR, "swh.journal.replay")

    # Fill the destination storage from Kafka
    dst = get_storage(cls="memory")
    worker_fn = functools.partial(process_replay_objects, storage=dst)
    nb_inserted = replayer.process(worker_fn)
    assert nb_sent == nb_inserted

    # check the logs for the collision being properly detected
    nb_collisions = 0
    for record in caplog.records:
        logtext = record.getMessage()
        if "Collision detected:" in logtext:
            nb_collisions += 1

    assert nb_collisions == 0, "No collision should be detected"

    # all objects from the src should exists in the dst storage, plus two more
    assert isinstance(src, InMemoryStorage)  # needed to help mypy
    assert isinstance(dst, InMemoryStorage)  # needed to help mypy

    check_replayed(src, dst, exclude=["contents"])

    assert (
        len(list(dst._cql_runner._contents.iter_all()))
        == len(list(src._cql_runner._contents.iter_all())) + 2
    )


def test_replay_skipped_content(replayer_storage_and_client):
    """Test the 'skipped_content' topic is properly replayed."""
    src, replayer = replayer_storage_and_client
    _check_replay_skipped_content(src, replayer, "skipped_content")


# utility functions


def check_replayed(
    src: InMemoryStorage,
    dst: InMemoryStorage,
    exclude: Optional[Container] = None,
    expected_anonymized=False,
):
    """Simple utility function to compare the content of 2 in_memory storages"""

    def fix_expected(attr, row):
        if expected_anonymized:
            if attr == "releases":
                row = dataclasses.replace(
                    row, author=row.author and row.author.anonymize()
                )
            elif attr == "revisions":
                row = dataclasses.replace(
                    row,
                    author=row.author.anonymize(),
                    committer=row.committer.anonymize(),
                )
        if attr == "revisions":
            # the replayer should now drop the metadata attribute; see
            # swh/storgae/replay.py:_insert_objects()
            row.metadata = "null"

        return row

    for attr_ in (
        "contents",
        "skipped_contents",
        "directories",
        "extid",
        "revisions",
        "releases",
        "snapshots",
        "origins",
        "origin_visits",
        "origin_visit_statuses",
        "raw_extrinsic_metadata",
    ):
        if exclude and attr_ in exclude:
            continue
        expected_objects = [
            (id, nullify_ctime(fix_expected(attr_, obj)))
            for id, obj in sorted(getattr(src._cql_runner, f"_{attr_}").iter_all())
        ]
        got_objects = [
            (id, nullify_ctime(obj))
            for id, obj in sorted(getattr(dst._cql_runner, f"_{attr_}").iter_all())
        ]
        assert got_objects == expected_objects, f"Mismatch object list for {attr_}"


def _check_replay_skipped_content(storage, replayer, topic):
    skipped_contents = _gen_skipped_contents(100)
    nb_sent = len(skipped_contents)
    producer = storage.journal_writer.journal.producer
    prefix = storage.journal_writer.journal._prefix

    for i, obj in enumerate(skipped_contents):
        content = SkippedContent.from_dict(obj)
        producer.produce(
            topic=f"{prefix}.{topic}",
            key=key_to_kafka(content.unique_key()),
            value=value_to_kafka(content.to_dict()),
        )
    producer.flush()

    dst_storage = get_storage(cls="memory")
    worker_fn = functools.partial(process_replay_objects, storage=dst_storage)
    nb_inserted = replayer.process(worker_fn)

    assert nb_sent == nb_inserted
    for content in skipped_contents:
        assert not storage.content_find({"sha1": content["sha1"]})

    # no skipped_content_find API endpoint, so use this instead
    assert not list(dst_storage.skipped_content_missing(skipped_contents))


def _updated(d1, d2):
    d1.update(d2)
    d1.pop("data", None)
    return d1


def _gen_skipped_contents(n=10):
    # we do not use the hypothesis strategy here because this does not play well with
    # pytest fixtures, and it makes test execution very slow
    now = datetime.datetime.now(tz=UTC)
    return [
        _updated(
            MultiHash.from_data(data=f"foo{i}".encode()).digest(),
            {
                "status": "absent",
                "reason": "why not",
                "origin": f"https://somewhere/{i}",
                "ctime": now,
                "length": len(f"foo{i}"),
            },
        )
        for i in range(n)
    ]


@pytest.mark.parametrize("privileged", [True, False])
def test_storage_replay_anonymized(
    kafka_prefix: str,
    kafka_consumer_group: str,
    kafka_server: str,
    privileged: bool,
):
    """Optimal replayer scenario.

    This:
    - writes objects to the topic
    - replayer consumes objects from the topic and replay them

    This tests the behavior with both a privileged and non-privileged replayer
    """
    writer_config = {
        "cls": "kafka",
        "brokers": [kafka_server],
        "client_id": "kafka_writer",
        "prefix": kafka_prefix,
        "anonymize": True,
        "auto_flush": False,
    }
    src_config: Dict[str, Any] = {"cls": "memory", "journal_writer": writer_config}

    storage = get_storage(**src_config)

    # Fill the src storage
    nb_sent = 0
    for obj_type, objs in TEST_OBJECTS.items():
        if obj_type in (
            OriginVisit.object_type,
            OriginVisitStatus.object_type,
        ):
            # these are unrelated with what we want to test here
            continue
        method = getattr(storage, f"{obj_type}_add")
        method(objs)
        nb_sent += len(objs)
    journal: JournalWriterInterface = storage.journal_writer.journal  # type:ignore
    journal.flush()

    # Fill a destination storage from Kafka, potentially using privileged topics
    dst_storage = get_storage(cls="memory")
    deserializer = ModelObjectDeserializer(
        validate=False
    )  # we cannot validate an anonymized replay
    replayer = JournalClient(
        brokers=kafka_server,
        group_id=kafka_consumer_group,
        prefix=kafka_prefix,
        stop_after_objects=nb_sent,
        privileged=privileged,
        value_deserializer=deserializer.convert,
    )
    worker_fn = functools.partial(process_replay_objects, storage=dst_storage)

    nb_inserted = replayer.process(worker_fn)
    replayer.consumer.commit()

    assert nb_sent == nb_inserted
    # Check the contents of the destination storage, and whether the anonymization was
    # properly used
    assert isinstance(storage, InMemoryStorage)  # needed to help mypy
    assert isinstance(dst_storage, InMemoryStorage)
    check_replayed(storage, dst_storage, expected_anonymized=not privileged)


def test_storage_replayer_with_validation_ok(
    replayer_storage_and_client, caplog, redisdb
):
    """Optimal replayer scenario

    with validation activated and reporter set to a redis db.

    - writes objects to a source storage
    - replayer consumes objects from the topic and replays them
    - a destination storage is filled from this
    - nothing has been reported in the redis db
    - both storages should have the same content
    """
    src, replayer = replayer_storage_and_client
    replayer.deserializer = ModelObjectDeserializer(validate=True, reporter=redisdb.set)

    # Fill Kafka using a source storage
    nb_sent = 0
    for object_type, objects in TEST_OBJECTS.items():
        method = getattr(src, f"{object_type}_add")
        method(objects)
        nb_sent += len(objects)
    src.journal_writer.journal.flush()

    # Fill the destination storage from Kafka
    dst = get_storage(cls="memory")
    worker_fn = functools.partial(process_replay_objects, storage=dst)
    nb_inserted = replayer.process(worker_fn)
    assert nb_sent == nb_inserted

    # check we do not have invalid objects reported
    invalid = 0
    for record in caplog.records:
        logtext = record.getMessage()
        if WRONG_ID_REG.match(logtext):
            invalid += 1
    assert invalid == 0, "Invalid objects should not be detected"
    assert not redisdb.keys()
    # so the dst should be the same as src storage
    check_replayed(cast(InMemoryStorage, src), cast(InMemoryStorage, dst))


def test_storage_replayer_with_validation_nok(
    replayer_storage_and_client, caplog, redisdb
):
    """Replayer scenario with invalid objects

    with validation and reporter set to a redis db.

    - writes objects to a source storage
    - replayer consumes objects from the topic and replays them
    - the destination storage is filled with only valid objects
    - the redis db contains the invalid (raw kafka mesg) objects
    """
    src, replayer = replayer_storage_and_client
    replayer.value_deserializer = ModelObjectDeserializer(
        validate=True, reporter=redisdb.set
    ).convert

    caplog.set_level(logging.ERROR, "swh.journal.replay")

    # Fill Kafka using a source storage
    nb_sent = 0
    for object_type, objects in TEST_OBJECTS.items():
        method = getattr(src, f"{object_type}_add")
        method(objects)
        nb_sent += len(objects)

    # insert invalid objects
    for object_type in (
        Revision.object_type,
        Directory.object_type,
        Release.object_type,
        Snapshot.object_type,
    ):
        method = getattr(src, f"{object_type}_add")
        method([attr.evolve(TEST_OBJECTS[object_type][0], id=b"\x00" * 20)])
        nb_sent += 1
    # also add an object that won't even be possible to instantiate; this needs
    # to be done at low kafka level (since we cannot instantiate the invalid model
    # object...)
    # we use directory[1] because it actually have some entries
    dict_repr = {
        # copy each dir entry twice
        "entries": TEST_OBJECTS[Directory.object_type][1].to_dict()["entries"] * 2,
        "id": b"\x01" * 20,
    }
    topic = f"{src.journal_writer.journal._prefix}.directory"
    src.journal_writer.journal.send(topic, dict_repr["id"], dict_repr)
    nb_sent += 1

    src.journal_writer.journal.flush()

    # Fill the destination storage from Kafka
    dst = get_storage(cls="tenacious", storage={"cls": "memory"})
    worker_fn = functools.partial(process_replay_objects, storage=dst)
    nb_inserted = replayer.process(worker_fn)
    assert nb_sent == nb_inserted

    # check we do have invalid objects reported
    invalid = 0
    for record in caplog.records:
        logtext = record.getMessage()
        if WRONG_ID_REG.match(logtext):
            invalid += 1
    assert invalid == 4, "Invalid objects should be detected"
    keys = validate_keys(redisdb.keys())

    assert {k for ts, k in keys} == {
        f"{typ}:{'0' * 40}" for typ in ("release", "revision", "snapshot", "directory")
    } | {"directory:" + "01" * 20}

    for key in redisdb.keys():
        # check the stored value looks right
        rawvalue = redisdb.get(key)
        error_struct = yaml.load(rawvalue, Loader=yaml.Loader)
        assert "kafka_message" in error_struct
        value = kafka_to_value(error_struct["kafka_message"])
        assert isinstance(value, dict)
        assert "id" in value
        assert value["id"] in (b"\x00" * 20, b"\x01" * 20)
        assert "obj" in error_struct
        assert "exc" in error_struct

    # check that invalid objects did not reach the dst storage
    for attr_ in (
        "directories",
        "revisions",
        "releases",
        "snapshots",
    ):
        for id, obj in sorted(getattr(dst._cql_runner, f"_{attr_}").iter_all()):
            assert id not in (b"\x00" * 20, b"\x01" * 20)

    # check that valid objects did reach the dst storage
    # revisions
    expected = [
        attr.evolve(rev, metadata=None) for rev in TEST_OBJECTS[Revision.object_type]
    ]
    result = dst.revision_get([obj.id for obj in TEST_OBJECTS[Revision.object_type]])
    assert result == expected
    # releases
    expected = TEST_OBJECTS[Release.object_type]
    result = dst.release_get([obj.id for obj in TEST_OBJECTS[Release.object_type]])
    assert result == expected
    # snapshot
    # result from snapshot_get is paginated, so adapt the expected to be comparable
    expected = [
        {"next_branch": None, **obj.to_dict()}
        for obj in TEST_OBJECTS[Snapshot.object_type]
    ]
    result = [dst.snapshot_get(obj.id) for obj in TEST_OBJECTS[Snapshot.object_type]]
    assert result == expected
    # directories
    for directory in TEST_OBJECTS[Directory.object_type]:
        assert set(dst.directory_get_entries(directory.id).results) == set(
            directory.entries
        )


def test_storage_replayer_with_validation_nok_with_exceptions(
    replayer_storage_and_client, caplog, redisdb
):
    """Replayer scenario with invalid objects, with exceptions

    with validation and reporter set to a redis db.

    - writes invalid objects to a source storage
    - replayer consumes objects from the topic and replays them (invalid
      objects being in the exception list)
    - the destination storage is filled with all objects
    - the redis db does not contain the invalid objects

    """
    src, replayer = replayer_storage_and_client

    caplog.set_level(logging.WARNING, "swh.journal.replay")

    # Fill Kafka using a source storage
    nb_sent = 0
    for object_type, objects in TEST_OBJECTS.items():
        method = getattr(src, f"{object_type}_add")
        method(objects)
        nb_sent += len(objects)

    # insert invalid objects
    known_invalid = []
    all_objs = {}
    for object_type in (
        Revision.object_type,
        Directory.object_type,
        Release.object_type,
        Snapshot.object_type,
    ):
        all_objs[object_type] = TEST_OBJECTS[object_type][:]
        method = getattr(src, f"{object_type}_add")
        invalid_obj = attr.evolve(TEST_OBJECTS[object_type][0], id=b"\x00" * 20)
        method([invalid_obj])
        nb_sent += 1
        all_objs[object_type].append(invalid_obj)
        known_invalid.append(
            (object_type.value, b"\x00" * 20, TEST_OBJECTS[object_type][0].id)
        )

    replayer.value_deserializer = ModelObjectDeserializer(
        validate=True,
        reporter=redisdb.set,
        known_mismatched_hashes=tuple(known_invalid),
    ).convert
    src.journal_writer.journal.flush()

    # Fill the destination storage from Kafka
    dst = get_storage(cls="memory")
    worker_fn = functools.partial(process_replay_objects, storage=dst)
    nb_inserted = replayer.process(worker_fn)
    assert nb_sent == nb_inserted

    # check we do have invalid but replicates objects reported
    invalid = 0
    for record in caplog.records:
        logtext = record.getMessage()
        if WRONG_ID_REG.match(logtext):
            assert logtext.endswith("Known exception, replicating it anyway.")
            invalid += 1
    assert invalid == 4, "Replayed invalid objects should be detected"
    assert not set(redisdb.keys())

    # check that all objects did reach the dst storage
    # revisions
    expected = [
        attr.evolve(rev, metadata=None) for rev in all_objs[Revision.object_type]
    ]
    result = dst.revision_get([obj.id for obj in all_objs[Revision.object_type]])
    assert result == expected
    # releases
    expected = all_objs[Release.object_type]
    result = dst.release_get([obj.id for obj in all_objs[Release.object_type]])
    assert result == expected
    # snapshot
    # result from snapshot_get is paginated, so adapt the expected to be comparable
    expected = [
        {"next_branch": None, **obj.to_dict()} for obj in all_objs[Snapshot.object_type]
    ]
    result = [dst.snapshot_get(obj.id) for obj in all_objs[Snapshot.object_type]]
    assert result == expected
    # directories
    for directory in all_objs[Directory.object_type]:
        assert set(dst.directory_get_entries(directory.id).results) == set(
            directory.entries
        )


def test_storage_replayer_with_validation_nok_raises(
    replayer_storage_and_client, caplog, redisdb
):
    """Replayer scenario with invalid objects

    with raise_on_error set to True

    This:
    - writes both valid & invalid objects to a source storage
    - a StorageArgumentException should be raised while replayer consumes
      objects from the topic and replays them
    """
    src, replayer = replayer_storage_and_client
    replayer.value_deserializer = ModelObjectDeserializer(
        validate=True, reporter=redisdb.set, raise_on_error=True
    ).convert

    caplog.set_level(logging.ERROR, "swh.journal.replay")

    # Fill Kafka using a source storage
    nb_sent = 0
    for object_type, objects in TEST_OBJECTS.items():
        method = getattr(src, f"{object_type}_add")
        method(objects)
        nb_sent += len(objects)

    # insert invalid objects
    for object_type in (
        Revision.object_type,
        Directory.object_type,
        Release.object_type,
        Snapshot.object_type,
    ):
        method = getattr(src, f"{object_type}_add")
        method([attr.evolve(TEST_OBJECTS[object_type][0], id=b"\x00" * 20)])
        nb_sent += 1

    # Fill the destination storage from Kafka
    dst = get_storage(cls="memory")
    worker_fn = functools.partial(process_replay_objects, storage=dst)
    with pytest.raises(StorageArgumentException):
        replayer.process(worker_fn)

    # check we do have invalid objects reported
    invalid = 0
    for record in caplog.records:
        logtext = record.getMessage()
        if WRONG_ID_REG.match(logtext):
            invalid += 1
    assert invalid == 1, "One invalid objects should be detected"
    assert len(redisdb.keys()) == 1


def test_storage_replayer_tenacious_with_hashcollisions(
    replayer_storage_and_client, caplog, redisdb
):
    """Replayer w/ tenacious proxy, filled with hash colliding content objects

    Ensure colliding objects are inserted properly
    """
    src, replayer = replayer_storage_and_client
    jwriter = src.journal_writer

    caplog.set_level(logging.ERROR, "swh.journal.replay")
    # Fill Kafka using the journal writer
    contents = [attr.evolve(c, ctime=now()) for c in TEST_OBJECTS["content"]]

    dst = get_storage(
        cls="tenacious", storage={"cls": "memory"}, error_reporter=redisdb.set
    )
    worker_fn = functools.partial(process_replay_objects, storage=dst)

    jwriter.content_add(contents)
    jwriter.journal.flush()
    nb_processed = replayer.process(worker_fn)
    assert nb_processed == len(contents)
    assert len(dst.storage._cql_runner._contents.data) == len(contents)
    assert not redisdb.keys()

    # hash collisions on sha256
    colliding_objs = [attr.evolve(c, sha256=b"\x00" * 32, data=None) for c in contents]
    jwriter.content_add(colliding_objs)
    jwriter.journal.flush()
    nb_processed = replayer.process(worker_fn)
    assert nb_processed == len(contents)
    assert sum(
        len(x) for x in dst.storage._cql_runner._contents.data.values()
    ) == 2 * len(contents)
    assert set(redisdb.keys()) == set()

    # hash collisions on blake2s256
    colliding_objs = [
        attr.evolve(c, blake2s256=b"\x00" * 32, data=None) for c in contents
    ]
    jwriter.content_add(colliding_objs)
    jwriter.journal.flush()
    nb_processed = replayer.process(worker_fn)
    assert nb_processed == len(contents)
    assert sum(
        len(x) for x in dst.storage._cql_runner._contents.data.values()
    ) == 3 * len(contents)
    assert set(redisdb.keys()) == set()

    # hash collisions on sha1
    colliding_objs = [attr.evolve(c, sha1=b"\x00" * 20, data=None) for c in contents]
    jwriter.content_add(colliding_objs)
    jwriter.journal.flush()
    nb_processed = replayer.process(worker_fn)
    assert nb_processed == len(contents)
    assert sum(
        len(x) for x in dst.storage._cql_runner._contents.data.values()
    ) == 4 * len(contents)
    assert set(redisdb.keys()) == set()

    # hash collisions on sha1_git
    colliding_objs = [
        attr.evolve(c, sha1_git=b"\x00" * 20, data=None) for c in contents
    ]
    jwriter.content_add(colliding_objs)
    jwriter.journal.flush()
    nb_processed = replayer.process(worker_fn)
    assert nb_processed == len(contents)
    assert sum(
        len(x) for x in dst.storage._cql_runner._contents.data.values()
    ) == 5 * len(contents)
    assert set(redisdb.keys()) == set()


def test_storage_replayer_tenacious_with_invalid_objects(
    replayer_storage, replayer_client, caplog, redisdb
):
    """Replayer scenario with invalid objects

    with raise_on_error set to True

    This:
    - writes both valid & invalid objects to a source storage
    - a StorageArgumentException should be raised while replayer consumes
      objects from the topic and replays them
    """
    # disable the validation made by the replayer itself since we want to test
    # errors occurring in the storage backend here
    replayer_client.value_deserializer.__self__.validate = False
    jwriter = replayer_storage.journal_writer

    caplog.set_level(logging.ERROR, "swh.journal.replay")

    n_inserted = 0
    # insert invalid objects
    for object_type in (
        Revision.object_type,
        Directory.object_type,
        Release.object_type,
        Snapshot.object_type,
    ):
        inv_obj = attr.evolve(TEST_OBJECTS[object_type][0], id=b"\x00" * 20)
        jwriter.write_additions(
            object_type,
            TEST_OBJECTS[object_type] + [inv_obj],
        )
        n_inserted += len(TEST_OBJECTS[object_type]) + 1
    jwriter.journal.flush()

    # Fill the destination storage from Kafka
    dst = get_storage(
        cls="tenacious",
        storage={"cls": "validate", "storage": {"cls": "memory"}},
        error_reporter=redisdb.set,
    )
    worker_fn = functools.partial(process_replay_objects, storage=dst)
    nb_processed = replayer_client.process(worker_fn)

    assert nb_processed == n_inserted
    assert len(redisdb.keys()) == 4

    object_types = []
    for ts, k in validate_keys(redisdb.keys()):
        object_types.append(k)

        # values are yaml serialized dicts with 2 keys: obj (as a dict) and exc
        # (dump of the exception as a list of str)
        value = redisdb.get(f"{ts}/{k}")
        # cannot use safe_load because this latter does not know how to load
        # tuples, which can be generated by some ModelObject.to_dict() implems
        error_struct = yaml.load(value, Loader=yaml.Loader)
        assert "obj" in error_struct
        assert error_struct["obj"]["id"] == b"\x00" * 20
        assert "exc" in error_struct
        assert "\n".join(error_struct["exc"]).startswith("Traceback")

    assert set(object_types) == {"revision", "directory", "release", "snapshot"}


def validate_keys(keys: List[bytes]) -> List[Tuple[str, str]]:
    """Check error reporter keys (from redis)

    Check they are correctly foratted and return them without the timestamp"""
    # keys are 'iso-date/remaiming'
    # check the timestamps for reported errors are in a reasonable range
    n = now()
    dt0 = datetime.timedelta(0)
    dt1 = datetime.timedelta(minutes=1)
    output = []
    for k in (x.decode() for x in keys):
        assert "/" in k
        tstamp, rem = k.split("/", 1)
        dt = n - datetime.datetime.fromisoformat(tstamp)
        assert dt0 < dt < dt1
        output.append((tstamp, rem))
    return output



# Copyright (C) 2018-2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

# Utility module to efficiently compute the list of changed files
# between two directory trees.
# The implementation is inspired from the work of Alberto Cortés
# for the go-git project. For more details, you can refer to:
#   - this blog post: https://blog.sourced.tech/post/difftree/
#   - the reference implementation in go:
#     https://github.com/src-d/go-git/tree/master/utils/merkletrie


import collections
from typing import Any, Dict, List, Optional

from swh.model.model import Directory
from swh.storage.interface import StorageInterface

from .dir_iterators import DirectoryIterator, DoubleDirectoryIterator, Remaining

# get the hash identifier for an empty directory
_empty_dir_hash = Directory(entries=()).id


def _get_rev(storage: StorageInterface, rev_id: bytes) -> Dict[str, Any]:
    """
    Return revision data from swh storage.
    """
    revision = storage.revision_get([rev_id])[0]
    assert revision is not None
    return revision.to_dict()


class _RevisionChangesList:
    """
    Helper class to track the changes between two
    revision directories.
    """

    def __init__(self, storage: StorageInterface, track_renaming: bool):
        """
        Args:
            storage: instance of swh storage
            track_renaming: whether to track or not files renaming
        """
        self.storage = storage
        self.track_renaming = track_renaming
        self.result: List[Dict[str, Any]] = []
        # dicts used to track file renaming based on hash value
        # we use a list instead of a single entry to handle the corner
        # case when a repository contains multiple instance of
        # the same file in different directories and a commit
        # renames all of them
        self.inserted_hash_idx: Dict[bytes, List[int]] = collections.defaultdict(list)
        self.deleted_hash_idx: Dict[bytes, List[int]] = collections.defaultdict(list)

    def add_insert(self, it_to: DirectoryIterator) -> None:
        """
        Add a file insertion in the to directory.

        Args:
            it_to: iterator on the to directory
        """
        to_hash = it_to.current_hash()
        # if the current file hash has been previously marked as deleted,
        # the file has been renamed
        if self.track_renaming and self.deleted_hash_idx[to_hash]:
            # pop the delete change index in the same order it was inserted
            change = self.result[self.deleted_hash_idx[to_hash].pop(0)]
            # change the delete change as a rename one
            change["type"] = "rename"
            change["to"] = it_to.current()
            change["to_path"] = it_to.current_path()
        else:
            # add the insert change in the list
            self.result.append(
                {
                    "type": "insert",
                    "from": None,
                    "from_path": None,
                    "to": it_to.current(),
                    "to_path": it_to.current_path(),
                }
            )
            # if rename tracking is activated, add the change index in
            # the inserted_hash_idx dict
            if self.track_renaming:
                self.inserted_hash_idx[to_hash].append(len(self.result) - 1)

    def add_delete(self, it_from: DirectoryIterator) -> None:
        """
        Add a file deletion in the from directory.

        Args:
            it_from: iterator on the from directory
        """
        from_hash = it_from.current_hash()
        # if the current file has been previously marked as inserted,
        # the file has been renamed
        if self.track_renaming and self.inserted_hash_idx[from_hash]:
            # pop the insert change index in the same order it was inserted
            change = self.result[self.inserted_hash_idx[from_hash].pop(0)]
            # change the insert change as a rename one
            change["type"] = "rename"
            change["from"] = it_from.current()
            change["from_path"] = it_from.current_path()
        else:
            # add the delete change in the list
            self.result.append(
                {
                    "type": "delete",
                    "from": it_from.current(),
                    "from_path": it_from.current_path(),
                    "to": None,
                    "to_path": None,
                }
            )
            # if rename tracking is activated, add the change index in
            # the deleted_hash_idx dict
            if self.track_renaming:
                self.deleted_hash_idx[from_hash].append(len(self.result) - 1)

    def add_modify(self, it_from: DirectoryIterator, it_to: DirectoryIterator) -> None:
        """
        Add a file modification in the to directory.

        Args:
            it_from: iterator on the from directory
            it_to: iterator on the to directory
        """
        self.result.append(
            {
                "type": "modify",
                "from": it_from.current(),
                "from_path": it_from.current_path(),
                "to": it_to.current(),
                "to_path": it_to.current_path(),
            }
        )

    def add_recursive(self, it: DirectoryIterator, insert: bool) -> None:
        """
        Recursively add changes from a directory.

        Args:
            it: iterator on a directory
            insert: the type of changes to add (insertion or deletion)
        """
        # current iterated element is a regular file,
        # simply add adequate change in the list
        if not it.current_is_dir():
            if insert:
                self.add_insert(it)
            else:
                self.add_delete(it)
            return
        # current iterated element is a directory,
        dir_id = it.current_hash()
        # handle empty dir insertion/deletion as the swh model allow
        # to have such object compared to git
        if dir_id == _empty_dir_hash:
            if insert:
                self.add_insert(it)
            else:
                self.add_delete(it)
        # iterate on files reachable from it and add
        # adequate changes in the list
        else:
            current_path = it.current_path()
            assert current_path is not None
            sub_it = DirectoryIterator(self.storage, dir_id, current_path + b"/")
            sub_it_current = sub_it.step()
            while sub_it_current:
                if not sub_it.current_is_dir():
                    if insert:
                        self.add_insert(sub_it)
                    else:
                        self.add_delete(sub_it)
                sub_it_current = sub_it.step()

    def add_recursive_insert(self, it_to: DirectoryIterator) -> None:
        """
        Recursively add files insertion from a to directory.

        Args:
            it_to: iterator on a to directory
        """
        self.add_recursive(it_to, True)

    def add_recursive_delete(self, it_from: DirectoryIterator) -> None:
        """
        Recursively add files deletion from a from directory.

        Args:
            it_from: iterator on a from directory
        """
        self.add_recursive(it_from, False)


def _diff_elts_same_name(
    changes: _RevisionChangesList, it: DoubleDirectoryIterator
) -> None:
    """ "
    Compare two directory entries with the same name and add adequate
    changes if any.

    Args:
        changes: the list of changes between two revisions
        it: the iterator traversing two revision directories at the same time
    """
    # compare the two current directory elements of the iterator
    status = it.compare()
    # elements have same hash and same permissions:
    # no changes to add and call next on the two iterators
    if status["same_hash"] and status["same_perms"]:
        it.next_both()
    # elements are regular files and have been modified:
    # insert the modification change in the list and
    # call next on the two iterators
    elif status["both_are_files"]:
        changes.add_modify(it.it_from, it.it_to)
        it.next_both()
    # one element is a regular file, the other a directory:
    # recursively add delete/insert changes and call next
    # on the two iterators
    elif status["file_and_dir"]:
        changes.add_recursive_delete(it.it_from)
        changes.add_recursive_insert(it.it_to)
        it.next_both()
    # both elements are directories:
    elif status["both_are_dirs"]:
        # from directory is empty:
        # recursively add insert changes in the to directory
        # and call next on the two iterators
        if status["from_is_empty_dir"]:
            changes.add_recursive_insert(it.it_to)
            it.next_both()
        # to directory is empty:
        # recursively add delete changes in the from directory
        # and call next on the two iterators
        elif status["to_is_empty_dir"]:
            changes.add_recursive_delete(it.it_from)
            it.next_both()
        # both directories are not empty:
        # call step on the two iterators to descend further in
        # the directory trees.
        elif not status["from_is_empty_dir"] and not status["to_is_empty_dir"]:
            it.step_both()


def _compare_paths(path1: bytes, path2: bytes) -> int:
    """
    Compare paths in lexicographic depth-first order.
    For instance, it returns:

        - "a" < "b"
        - "b/c/d" < "b"
        - "c/foo.txt" < "c.txt"
    """
    path1_parts = path1.split(b"/")
    path2_parts = path2.split(b"/")
    i = 0
    while True:
        if len(path1_parts) == len(path2_parts) and i == len(path1_parts):
            return 0
        elif len(path2_parts) == i:
            return 1
        elif len(path1_parts) == i:
            return -1
        else:
            if path2_parts[i] > path1_parts[i]:
                return -1
            elif path2_parts[i] < path1_parts[i]:
                return 1
        i = i + 1


def _diff_elts(changes: _RevisionChangesList, it: DoubleDirectoryIterator) -> None:
    """
    Compare two directory entries.

    Args:
        changes: the list of changes between two revisions
        it: the iterator traversing two revision directories at the same time
    """
    from_current_path = it.it_from.current_path()
    to_current_path = it.it_to.current_path()
    assert from_current_path is not None
    assert to_current_path is not None
    # compare current to and from path in depth-first lexicographic order
    c = _compare_paths(from_current_path, to_current_path)
    # current from path is lower than the current to path:
    # the from path has been deleted
    if c < 0:
        changes.add_recursive_delete(it.it_from)
        it.next_from()
    # current from path is greater than the current to path:
    # the to path has been inserted
    elif c > 0:
        changes.add_recursive_insert(it.it_to)
        it.next_to()
    # paths are the same and need more processing
    else:
        _diff_elts_same_name(changes, it)


def diff_directories(
    storage: StorageInterface,
    from_dir: Optional[bytes],
    to_dir: bytes,
    track_renaming: bool = False,
) -> List[Dict[str, Any]]:
    """
    Compute the differential between two directories, i.e. the list of
    file changes (insertion / deletion / modification / renaming)
    between them.

    Args:
        storage: instance of a swh storage (either local or remote, for optimal
            performance the use of a local storage is recommended)
        from_dir: the swh identifier of the directory to compare from
        to_dir: the swh identifier of the directory to compare to
        track_renaming: whether or not to track files renaming

    Returns:
        A list of dict representing the changes between the two
        revisions. Each dict contains the following entries:

            - ``type``: a string describing the type of change
              (``insert`` / ``delete`` / ``modify`` / ``rename``)

            - ``from``: a dict containing the directory entry metadata in the
              from revision (:const:`None` in case of an insertion)

            - ``from_path``: bytes string corresponding to the absolute path
              of the from revision entry (:const:`None` in case of an insertion)

            - ``to``: a dict containing the directory entry metadata in the
              to revision (:const:`None` in case of a deletion)

            - ``to_path``: bytes string corresponding to the absolute path
              of the to revision entry (:const:`None` in case of a deletion)

        The returned list is sorted in lexicographic depth-first order
        according to the value of the ``to_path`` field.

        .. warning::

            The algorithm used to track files renaming is quite naive (it compares hashes
            between deleted and inserted files) and might fail to detect all renamings for
            some edge cases.

    """
    changes = _RevisionChangesList(storage, track_renaming)
    it = DoubleDirectoryIterator(storage, from_dir, to_dir)
    while True:
        r = it.remaining()
        if r == Remaining.NoMoreFiles:
            break
        elif r == Remaining.OnlyFromFilesRemain:
            changes.add_recursive_delete(it.it_from)
            it.next_from()
        elif r == Remaining.OnlyToFilesRemain:
            changes.add_recursive_insert(it.it_to)
            it.next_to()
        else:
            _diff_elts(changes, it)
    return changes.result


def diff_revisions(
    storage: StorageInterface,
    from_rev: Optional[bytes],
    to_rev: bytes,
    track_renaming: bool = False,
) -> List[Dict[str, Any]]:
    """
    Compute the differential between two revisions,
    i.e. the list of file changes between the two associated directories.

    Args:
        storage: instance of a swh storage (either local or remote, for optimal
            performance the use of a local storage is recommended)
        from_rev: the identifier of the revision to compare from
        to_rev: the identifier of the revision to compare to
        track_renaming: whether or not to track files renaming

    Returns:
        A list of dict describing the introduced file changes
        (see :func:`swh.storage.algos.diff.diff_directories`).

    .. warning::

            The algorithm used to track files renaming is quite naive (it compares hashes
            between deleted and inserted files) and might fail to detect all renamings for
            some edge cases.

    """
    from_dir = None
    if from_rev is not None:
        from_dir = _get_rev(storage, from_rev)["directory"]
    to_dir = _get_rev(storage, to_rev)["directory"]
    return diff_directories(storage, from_dir, to_dir, track_renaming)


def diff_revision(
    storage: StorageInterface, revision: bytes, track_renaming: bool = False
) -> List[Dict[str, Any]]:
    """
    Computes the differential between a revision and its first parent.
    If the revision has no parents, the directory to compare from
    is considered as empty.
    In other words, it computes the file changes introduced in a
    specific revision.

    Args:
        storage: instance of a swh storage (either local or remote, for
            optimal performance the use of a local storage is recommended)
        revision: the identifier of the revision from which to compute the
            introduced changes.
        track_renaming: whether or not to track files renaming

    Returns:
        A list of dict describing the introduced file changes
        (see :func:`swh.storage.algos.diff.diff_directories`).

    .. warning::

            The algorithm used to track files renaming is quite naive (it compares hashes
            between deleted and inserted files) and might fail to detect all renamings for
            some edge cases.
    """
    rev_data = _get_rev(storage, revision)
    parent = None
    if rev_data["parents"]:
        parent = rev_data["parents"][0]
    return diff_revisions(storage, parent, revision, track_renaming)



# Copyright (C) 2024 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
from typing import Iterable, cast

import attr
import pytest

from swh.core.api.classes import stream_results
from swh.model.model import (
    Directory,
    DirectoryEntry,
    ExtID,
    OriginVisit,
    OriginVisitStatus,
    Release,
    ReleaseTargetType,
    Revision,
    RevisionType,
    Snapshot,
    SnapshotBranch,
    SnapshotTargetType,
)
from swh.model.swhids import CoreSWHID, ExtendedObjectType, ExtendedSWHID, ObjectType
from swh.storage.exc import MaskedObjectException
from swh.storage.interface import HashDict
from swh.storage.proxies.masking.cls import MaskingProxyStorage
from swh.storage.proxies.masking.db import MaskedState
from swh.storage.tests.storage_data import StorageData


def now() -> datetime.datetime:
    return datetime.datetime.now(tz=datetime.timezone.utc)


@pytest.fixture
def set_object_visibility(masking_admin):
    # Create a request
    request = masking_admin.create_request(slug="foo", reason="bar")

    def set_visibility(objects: Iterable[ExtendedSWHID], new_state: MaskedState):
        masking_admin.set_object_state(
            request_id=request.id,
            new_state=new_state,
            swhids=list(objects),
        )

    return set_visibility


def assert_masked_objects_raise(func, objects, set_object_visibility):
    """Check that the called function raises a MaskedObjectException for the
    list of objects, only if they're masked.

    To do so, we call the function three times:
      - once to record the expected return value for the function
      - once to check the exception is properly raised after masking the object
      - one last time check that the original behavior is restored when the
        object is unmasked.

    We expect that func's return value is deterministic, and non-falsey.
    """
    set_object_visibility(list(objects), MaskedState.VISIBLE)

    expected_result = func()

    assert expected_result

    set_object_visibility(list(objects), MaskedState.DECISION_PENDING)

    with pytest.raises(MaskedObjectException) as exc_info:
        func()
    assert exc_info.value.masked.keys() == set(objects)

    set_object_visibility(list(objects), MaskedState.VISIBLE)
    assert func() == expected_result

    return expected_result


def test_content_get_data(swh_storage, set_object_visibility):
    # Ensure that object masking doesn't prevent insertion
    set_object_visibility(
        [StorageData.content.swhid().to_extended()], MaskedState.DECISION_PENDING
    )

    swh_storage.content_add([StorageData.content, StorageData.content2])

    assert (
        swh_storage.content_get_data(StorageData.content2.hashes())
        == StorageData.content2.data
    )

    assert_masked_objects_raise(
        lambda: swh_storage.content_get_data(StorageData.content.hashes()),
        [StorageData.content.swhid().to_extended()],
        set_object_visibility,
    )

    with pytest.deprecated_call():
        assert_masked_objects_raise(
            lambda: swh_storage.content_get_data(StorageData.content.sha1),
            [StorageData.content.swhid().to_extended()],
            set_object_visibility,
        )


def test_content_get(swh_storage, set_object_visibility):
    # Ensure that object masking doesn't prevent insertion
    set_object_visibility(
        [StorageData.content.swhid().to_extended()], MaskedState.DECISION_PENDING
    )

    swh_storage.content_add([StorageData.content, StorageData.content2])

    assert swh_storage.content_get([StorageData.content2.sha256], algo="sha256") == [
        attr.evolve(StorageData.content2, data=None)
    ]

    assert_masked_objects_raise(
        lambda: swh_storage.content_get(
            [StorageData.content.sha256, StorageData.content2.sha256], algo="sha256"
        ),
        [StorageData.content.swhid().to_extended()],
        set_object_visibility,
    )


def test_content_find(swh_storage, set_object_visibility):
    # Ensure that object masking doesn't prevent insertion
    set_object_visibility(
        [StorageData.content.swhid().to_extended()], MaskedState.DECISION_PENDING
    )

    swh_storage.content_add([StorageData.content, StorageData.content3])

    assert swh_storage.content_find(StorageData.content3.hashes()) == [
        attr.evolve(StorageData.content3, data=None)
    ]

    assert_masked_objects_raise(
        lambda: swh_storage.content_find(StorageData.content.hashes()),
        [StorageData.content.swhid().to_extended()],
        set_object_visibility,
    )


def test_directory_ls(swh_storage, set_object_visibility):
    set_object_visibility(
        [StorageData.directory.swhid().to_extended()], MaskedState.DECISION_PENDING
    )

    swh_storage.directory_add([StorageData.directory, StorageData.directory2])

    dir2_ls = list(swh_storage.directory_ls(StorageData.directory2.id, recursive=False))
    assert dir2_ls

    assert_masked_objects_raise(
        lambda: list(swh_storage.directory_ls(StorageData.directory.id)),
        [StorageData.directory.swhid().to_extended()],
        set_object_visibility,
    )

    # Check that a masked entry doesn't affect the visibility of the directory itself

    assert StorageData.directory2.entries[0].type == "file"
    set_object_visibility(
        [
            ExtendedSWHID(
                object_id=StorageData.directory2.entries[0].target,
                object_type=ExtendedObjectType.CONTENT,
            )
        ],
        MaskedState.DECISION_PENDING,
    )

    assert (
        list(swh_storage.directory_ls(StorageData.directory2.id, recursive=False))
        == dir2_ls
    )


def test_directory_get_entries(swh_storage, set_object_visibility):
    set_object_visibility(
        [StorageData.directory.swhid().to_extended()], MaskedState.DECISION_PENDING
    )

    swh_storage.directory_add([StorageData.directory, StorageData.directory2])

    dir2_entries = list(
        stream_results(swh_storage.directory_get_entries, StorageData.directory2.id)
    )
    assert dir2_entries

    assert_masked_objects_raise(
        lambda: list(
            stream_results(swh_storage.directory_get_entries, StorageData.directory.id)
        ),
        [StorageData.directory.swhid().to_extended()],
        set_object_visibility,
    )

    # Check that a masked entry doesn't affect the visibility of the directory itself

    assert StorageData.directory2.entries[0].type == "file"
    set_object_visibility(
        [
            ExtendedSWHID(
                object_id=StorageData.directory2.entries[0].target,
                object_type=ExtendedObjectType.CONTENT,
            )
        ],
        MaskedState.DECISION_PENDING,
    )

    assert (
        list(
            stream_results(
                swh_storage.directory_get_entries,
                StorageData.directory2.id,
            )
        )
        == dir2_entries
    )


def test_directory_entry_get_by_path(swh_storage, set_object_visibility):
    set_object_visibility(
        [StorageData.directory.swhid().to_extended()], MaskedState.DECISION_PENDING
    )

    swh_storage.directory_add([StorageData.directory, StorageData.directory2])

    dir2_entries = swh_storage.directory_entry_get_by_path(
        StorageData.directory2.id,
        [StorageData.directory2.entries[0].name],
    )
    assert dir2_entries

    assert_masked_objects_raise(
        lambda: swh_storage.directory_entry_get_by_path(
            StorageData.directory.id,
            [StorageData.directory.entries[0].name],
        ),
        [StorageData.directory.swhid().to_extended()],
        set_object_visibility,
    )

    # Check that a masked entry doesn't affect the visibility of the directory itself

    assert StorageData.directory2.entries[0].type == "file"
    set_object_visibility(
        [
            ExtendedSWHID(
                object_id=StorageData.directory2.entries[0].target,
                object_type=ExtendedObjectType.CONTENT,
            )
        ],
        MaskedState.DECISION_PENDING,
    )

    assert (
        swh_storage.directory_entry_get_by_path(
            StorageData.directory2.id,
            [StorageData.directory2.entries[0].name],
        )
        == dir2_entries
    )


def test_directory_get_raw_manifest(swh_storage, set_object_visibility):
    directory_with_raw_manifest = StorageData.directory6
    assert directory_with_raw_manifest.raw_manifest

    set_object_visibility(
        [directory_with_raw_manifest.swhid().to_extended()],
        MaskedState.DECISION_PENDING,
    )

    swh_storage.directory_add([directory_with_raw_manifest, StorageData.directory2])

    assert swh_storage.directory_get_raw_manifest([StorageData.directory2.id]) == {
        StorageData.directory2.id: None
    }

    assert_masked_objects_raise(
        lambda: swh_storage.directory_get_raw_manifest(
            [directory_with_raw_manifest.id, StorageData.directory2.id],
        ),
        [directory_with_raw_manifest.swhid().to_extended()],
        set_object_visibility,
    )


def test_directory_get_id_partition(swh_storage, set_object_visibility):
    directories = [
        Directory(
            entries=(
                DirectoryEntry(
                    name=f"entry{i}".encode(),
                    type="file",
                    target=b"\x00" * 20,
                    perms=0,
                ),
            )
        )
        for i in range(100)
    ]
    swh_storage.directory_add(directories)

    nb_partitions = 4

    for partition in range(nb_partitions):
        if directories[0].id in stream_results(
            swh_storage.directory_get_id_partition, partition, nb_partitions
        ):
            break
    else:
        assert False, "No partition with directories[0] found?"

    assert_masked_objects_raise(
        lambda: list(
            stream_results(
                swh_storage.directory_get_id_partition, partition, nb_partitions
            )
        ),
        [directories[0].swhid().to_extended()],
        set_object_visibility,
    )


def test_revision_get(swh_storage, set_object_visibility):
    # revision2 has one parent, revision
    assert StorageData.revision2.parents == (StorageData.revision.id,)

    # Ensure that object masking doesn't prevent insertion
    set_object_visibility(
        [StorageData.revision.swhid().to_extended()], MaskedState.DECISION_PENDING
    )

    swh_storage.revision_add([StorageData.revision, StorageData.revision2])

    # This also checks that masking a parent doesn't mask the revision itself
    assert swh_storage.revision_get([StorageData.revision2.id]) == [
        StorageData.revision2
    ]

    # Check that masking the directory doesn't prevent getting the revision
    set_object_visibility(
        [
            ExtendedSWHID(
                object_id=StorageData.revision2.directory,
                object_type=ExtendedObjectType.DIRECTORY,
            )
        ],
        MaskedState.DECISION_PENDING,
    )
    assert swh_storage.revision_get([StorageData.revision2.id]) == [
        StorageData.revision2
    ]

    assert_masked_objects_raise(
        lambda: swh_storage.revision_get(
            [StorageData.revision.id, StorageData.revision2.id]
        ),
        [StorageData.revision.swhid().to_extended()],
        set_object_visibility,
    )


def test_revision_log(swh_storage, set_object_visibility):
    # revision2 has one parent, revision
    assert StorageData.revision2.parents == (StorageData.revision.id,)

    swh_storage.revision_add([StorageData.revision, StorageData.revision2])

    # Only mask the parent revision
    set_object_visibility(
        [StorageData.revision.swhid().to_extended()], MaskedState.DECISION_PENDING
    )

    # Getting one entry in the log still works
    assert list(swh_storage.revision_log([StorageData.revision2.id], limit=1))

    # But the parent is properly masked
    assert_masked_objects_raise(
        lambda: list(swh_storage.revision_log([StorageData.revision2.id], limit=2)),
        [StorageData.revision.swhid().to_extended()],
        set_object_visibility,
    )


def test_revision_shortlog(swh_storage, set_object_visibility):
    # revision2 has one parent, revision
    assert StorageData.revision2.parents == (StorageData.revision.id,)

    swh_storage.revision_add([StorageData.revision, StorageData.revision2])

    # Only mask the parent revision
    set_object_visibility(
        [StorageData.revision.swhid().to_extended()], MaskedState.DECISION_PENDING
    )

    # Getting one entry in the log still works
    assert list(swh_storage.revision_shortlog([StorageData.revision2.id], limit=1))

    # But the parent is properly masked
    assert_masked_objects_raise(
        lambda: list(
            swh_storage.revision_shortlog([StorageData.revision2.id], limit=2)
        ),
        [StorageData.revision.swhid().to_extended()],
        set_object_visibility,
    )


def test_revision_get_partition(swh_storage, set_object_visibility):
    revisions = [
        Revision(
            message=f"hello{i}".encode(),
            author=None,
            date=None,
            committer=None,
            committer_date=None,
            parents=(),
            type=RevisionType.GIT,
            directory=b"\x00" * 20,
            synthetic=True,
        )
        for i in range(100)
    ]
    swh_storage.revision_add(revisions)

    nb_partitions = 4

    for partition in range(nb_partitions):
        if revisions[0] in stream_results(
            swh_storage.revision_get_partition, partition, nb_partitions
        ):
            break
    else:
        assert False, "No partition with revisions[0] found?"

    assert_masked_objects_raise(
        lambda: list(
            stream_results(swh_storage.revision_get_partition, partition, nb_partitions)
        ),
        [revisions[0].swhid().to_extended()],
        set_object_visibility,
    )


def test_extid_get_from_target(swh_storage, set_object_visibility):
    gitids = [
        revision.id
        for revision in StorageData.revisions
        if revision.type.value == "git"
    ]
    extids = [
        ExtID(
            extid=gitid,
            extid_type="git",
            target=CoreSWHID(
                object_id=gitid,
                object_type=ObjectType.REVISION,
            ),
        )
        for gitid in gitids
    ]

    # Once more, check that visibility of both target and extid itself doesn't
    # affect insertion
    set_object_visibility(
        [extids[0].target.to_extended()], MaskedState.DECISION_PENDING
    )

    swh_storage.extid_add(extids)

    # Check that extids get masked by target
    assert_masked_objects_raise(
        lambda: swh_storage.extid_get_from_target(
            extids[0].target.object_type,
            [extids[0].target.object_id, extids[1].target.object_id],
        ),
        [extids[0].target.to_extended()],
        set_object_visibility,
    )


def test_extid_get_from_extid():
    # Should this mask extids if the target is masked?
    pass


def test_release_get(swh_storage, set_object_visibility):
    # release.target = revision
    assert StorageData.release.target == StorageData.revision.id
    assert StorageData.release.target_type == ReleaseTargetType.REVISION

    # release2.target = revision2
    assert StorageData.release2.target == StorageData.revision2.id
    assert StorageData.release2.target_type == ReleaseTargetType.REVISION

    # Ensure that object masking doesn't prevent insertion
    set_object_visibility(
        [StorageData.release.swhid().to_extended()], MaskedState.DECISION_PENDING
    )

    swh_storage.release_add([StorageData.release, StorageData.release2])

    assert swh_storage.release_get([StorageData.release2.id]) == [StorageData.release2]

    # Check that masking the target doesn't prevent getting the release
    set_object_visibility(
        [StorageData.revision2.swhid().to_extended()],
        MaskedState.DECISION_PENDING,
    )
    assert swh_storage.release_get([StorageData.release2.id]) == [StorageData.release2]

    assert_masked_objects_raise(
        lambda: swh_storage.release_get(
            [StorageData.release.id, StorageData.release2.id]
        ),
        [StorageData.release.swhid().to_extended()],
        set_object_visibility,
    )


def test_release_get_partition(swh_storage, set_object_visibility):
    releases = [
        Release(
            name=f"release{i}".encode(),
            message=f"hello{i}".encode(),
            author=None,
            date=None,
            target=b"\x00" * 20,
            target_type=ReleaseTargetType.REVISION,
            synthetic=True,
        )
        for i in range(100)
    ]
    swh_storage.release_add(releases)

    nb_partitions = 4

    for partition in range(nb_partitions):
        if releases[0] in stream_results(
            swh_storage.release_get_partition, partition, nb_partitions
        ):
            break
    else:
        assert False, "No partition with releases[0] found?"

    assert_masked_objects_raise(
        lambda: list(
            stream_results(swh_storage.release_get_partition, partition, nb_partitions)
        ),
        [releases[0].swhid().to_extended()],
        set_object_visibility,
    )


def test_snapshot_get(swh_storage, set_object_visibility):
    # Ensure that object masking doesn't prevent insertion
    set_object_visibility(
        [StorageData.snapshot.swhid().to_extended()], MaskedState.DECISION_PENDING
    )

    swh_storage.snapshot_add([StorageData.snapshot, StorageData.complete_snapshot])

    complete_snapshot_dict = {
        **StorageData.complete_snapshot.to_dict(),
        "next_branch": None,
    }

    assert (
        swh_storage.snapshot_get(StorageData.complete_snapshot.id)
        == complete_snapshot_dict
    )
    # Check that masking branch targets doesn't affect snapshot visibility
    for branch in StorageData.complete_snapshot.branches.values():
        if not branch or branch.target_type == SnapshotTargetType.ALIAS:
            continue

        set_object_visibility(
            [branch.swhid().to_extended()], MaskedState.DECISION_PENDING
        )
        assert (
            swh_storage.snapshot_get(StorageData.complete_snapshot.id)
            == complete_snapshot_dict
        )

    assert_masked_objects_raise(
        lambda: swh_storage.snapshot_get(StorageData.snapshot.id),
        [StorageData.snapshot.swhid().to_extended()],
        set_object_visibility,
    )


def test_snapshot_count_branches(swh_storage, set_object_visibility):
    swh_storage.snapshot_add([StorageData.snapshot, StorageData.complete_snapshot])

    assert_masked_objects_raise(
        lambda: swh_storage.snapshot_count_branches(StorageData.complete_snapshot.id),
        [StorageData.complete_snapshot.swhid().to_extended()],
        set_object_visibility,
    )


def test_snapshot_get_branches(swh_storage, set_object_visibility):
    set_object_visibility(
        [StorageData.snapshot.swhid().to_extended()], MaskedState.DECISION_PENDING
    )

    swh_storage.snapshot_add([StorageData.snapshot, StorageData.complete_snapshot])

    get_branches_result = {
        "id": StorageData.complete_snapshot.id,
        "branches": StorageData.complete_snapshot.branches,
        "next_branch": None,
    }

    assert (
        swh_storage.snapshot_get_branches(StorageData.complete_snapshot.id)
        == get_branches_result
    )
    # Check that masking branch targets doesn't affect snapshot visibility
    for branch in StorageData.complete_snapshot.branches.values():
        if not branch or branch.target_type == SnapshotTargetType.ALIAS:
            continue

        set_object_visibility(
            [branch.swhid().to_extended()], MaskedState.DECISION_PENDING
        )
        assert (
            swh_storage.snapshot_get_branches(StorageData.complete_snapshot.id)
            == get_branches_result
        )

    assert_masked_objects_raise(
        lambda: swh_storage.snapshot_get_branches(StorageData.snapshot.id),
        [StorageData.snapshot.swhid().to_extended()],
        set_object_visibility,
    )


def test_snapshot_branch_get_by_name(swh_storage, set_object_visibility):
    set_object_visibility(
        [StorageData.snapshot.swhid().to_extended()], MaskedState.DECISION_PENDING
    )

    swh_storage.snapshot_add([StorageData.snapshot, StorageData.complete_snapshot])

    assert swh_storage.snapshot_branch_get_by_name(
        StorageData.complete_snapshot.id, branch_name=b"content"
    ).branch_found

    # Check that masking branch targets doesn't affect snapshot visibility
    for name, branch in StorageData.complete_snapshot.branches.items():
        if branch and branch.target_type != SnapshotTargetType.ALIAS:
            set_object_visibility(
                [branch.swhid().to_extended()], MaskedState.DECISION_PENDING
            )

        found = swh_storage.snapshot_branch_get_by_name(
            StorageData.complete_snapshot.id, branch_name=name, follow_alias_chain=False
        )

        assert found.branch_found
        assert found.target == branch

    assert_masked_objects_raise(
        lambda: swh_storage.snapshot_branch_get_by_name(
            StorageData.snapshot.id, b"HEAD"
        ),
        [StorageData.snapshot.swhid().to_extended()],
        set_object_visibility,
    )


def test_snapshot_get_id_partition(swh_storage, set_object_visibility):
    snapshots = [
        Snapshot(
            branches={
                f"branch{i}".encode(): SnapshotBranch(
                    target=b"\x00" * 20,
                    target_type=SnapshotTargetType.REVISION,
                ),
            },
        )
        for i in range(100)
    ]
    swh_storage.snapshot_add(snapshots)

    nb_partitions = 4

    for partition in range(nb_partitions):
        if snapshots[0].id in stream_results(
            swh_storage.snapshot_get_id_partition, partition, nb_partitions
        ):
            break
    else:
        assert False, "No partition with snapshots[0] found?"

    assert_masked_objects_raise(
        lambda: list(
            stream_results(
                swh_storage.snapshot_get_id_partition, partition, nb_partitions
            )
        ),
        [snapshots[0].swhid().to_extended()],
        set_object_visibility,
    )


def test_origin_get(swh_storage, set_object_visibility):
    set_object_visibility([StorageData.origin.swhid()], MaskedState.DECISION_PENDING)
    swh_storage.origin_add([StorageData.origin, StorageData.origin2])

    assert [StorageData.origin2] == swh_storage.origin_get([StorageData.origin2.url])

    assert_masked_objects_raise(
        lambda: swh_storage.origin_get([StorageData.origin.url]),
        [StorageData.origin.swhid()],
        set_object_visibility,
    )


def test_origin_get_by_sha1(swh_storage, set_object_visibility):
    set_object_visibility([StorageData.origin.swhid()], MaskedState.DECISION_PENDING)
    swh_storage.origin_add([StorageData.origin, StorageData.origin2])

    assert [{"url": StorageData.origin2.url}] == swh_storage.origin_get_by_sha1(
        [StorageData.origin2.id]
    )

    assert_masked_objects_raise(
        lambda: swh_storage.origin_get_by_sha1([StorageData.origin.id]),
        [StorageData.origin.swhid()],
        set_object_visibility,
    )


def test_origin_list(swh_storage, set_object_visibility):
    swh_storage.origin_add(StorageData.origins)

    res = list(stream_results(swh_storage.origin_list))
    assert len(res) == len(StorageData.origins)

    assert_masked_objects_raise(
        lambda: list(stream_results(swh_storage.origin_list)),
        [StorageData.origin.swhid()],
        set_object_visibility,
    )


def test_origin_search(swh_storage, set_object_visibility):
    swh_storage.origin_add(StorageData.origins)

    assert_masked_objects_raise(
        lambda: list(stream_results(swh_storage.origin_search, StorageData.origin.url)),
        [StorageData.origin.swhid()],
        set_object_visibility,
    )


def add_a_couple_visits(swh_storage):
    """Adds a couple of visits of one origin to swh_storage,
    returning the origin and the latest visit added."""
    origin = StorageData.origins[0]
    swh_storage.origin_add([origin])

    visit2_date = now()
    visit1_date = visit2_date - datetime.timedelta(seconds=3600)
    visit_type = "git"

    # set first visit to a null snapshot
    visit1 = swh_storage.origin_visit_add(
        [
            OriginVisit(
                origin=origin.url,
                date=visit1_date,
                type=visit_type,
            )
        ]
    )[0]
    swh_storage.origin_visit_status_add(
        [
            OriginVisitStatus(
                origin=origin.url,
                visit=visit1.visit,
                date=visit1_date,
                status="created",
                snapshot=None,
            )
        ]
    )

    # set second visit to a non-null snapshot
    visit2 = swh_storage.origin_visit_add(
        [
            OriginVisit(
                origin=origin.url,
                date=visit2_date,
                type=visit_type,
            )
        ]
    )[0]
    swh_storage.origin_visit_status_add(
        [
            OriginVisitStatus(
                origin=origin.url,
                visit=visit2.visit,
                date=visit2.date,
                status="created",
                snapshot=None,
            ),
            OriginVisitStatus(
                origin=origin.url,
                visit=visit2.visit,
                date=visit2.date + datetime.timedelta(seconds=60),
                status="full",
                snapshot=StorageData.snapshot.id,
            ),
        ]
    )

    return origin, visit2


def test_origin_snapshot_get_all(swh_storage, set_object_visibility):
    origin, _ = add_a_couple_visits(swh_storage)

    set_object_visibility([StorageData.snapshot.swhid()], MaskedState.DECISION_PENDING)
    assert swh_storage.origin_snapshot_get_all(origin.url) == [StorageData.snapshot.id]

    assert_masked_objects_raise(
        lambda: swh_storage.origin_snapshot_get_all(origin.url),
        [origin.swhid()],
        set_object_visibility,
    )


def test_origin_visit_behavior(swh_storage, set_object_visibility):
    origin, last_visit = add_a_couple_visits(swh_storage)

    for method in (
        "origin_visit_get",
        "origin_visit_get_latest",
        "origin_visit_get_with_statuses",
    ):
        assert_masked_objects_raise(
            lambda method=method: getattr(swh_storage, method)(origin.url),
            [origin.swhid()],
            set_object_visibility,
        )

    assert_masked_objects_raise(
        lambda: swh_storage.origin_visit_find_by_date(origin.url, last_visit.date),
        [origin.swhid()],
        set_object_visibility,
    )

    for method in (
        "origin_visit_get_by",
        "origin_visit_status_get",
        "origin_visit_status_get_latest",
    ):
        assert_masked_objects_raise(
            lambda method=method: getattr(swh_storage, method)(
                origin.url, last_visit.visit
            ),
            [origin.swhid()],
            set_object_visibility,
        )


def test_raw_extrinsic_metadata(swh_storage, set_object_visibility):
    origin = StorageData.origin
    origin_metadata = StorageData.origin_metadata

    authority1 = origin_metadata[0].authority
    authority2 = origin_metadata[-1].authority
    assert authority1 != authority2

    swh_storage.origin_add([origin])

    swh_storage.metadata_fetcher_add([m.fetcher for m in origin_metadata])
    swh_storage.metadata_authority_add([m.authority for m in origin_metadata])

    swh_storage.raw_extrinsic_metadata_add(origin_metadata)

    # The last metadata should not be masked
    assert swh_storage.raw_extrinsic_metadata_get(origin.swhid(), authority2).results

    # Masking both the origin and the metadata's swhid should mask the REMD object
    for masked_swhid in [origin.swhid(), origin_metadata[0].swhid()]:
        assert_masked_objects_raise(
            lambda: swh_storage.raw_extrinsic_metadata_get(origin.swhid(), authority1),
            [masked_swhid],
            set_object_visibility,
        )

        assert_masked_objects_raise(
            lambda: swh_storage.raw_extrinsic_metadata_get_by_ids(
                [origin_metadata[0].id]
            ),
            [masked_swhid],
            set_object_visibility,
        )

    # Authority information is only masked if the target is masked
    found_authorities = assert_masked_objects_raise(
        lambda: swh_storage.raw_extrinsic_metadata_get_authorities(origin.swhid()),
        [origin.swhid()],
        set_object_visibility,
    )

    set_object_visibility([origin_metadata[0].swhid()], MaskedState.DECISION_PENDING)

    assert (
        swh_storage.raw_extrinsic_metadata_get_authorities(origin.swhid())
        == found_authorities
    )


def test_proxy_overhead_metric(swh_storage: MaskingProxyStorage, mocker) -> None:
    timing = mocker.patch("swh.core.statsd.statsd.timing")
    monotonic = mocker.patch("swh.storage.metrics.monotonic")
    monotonic.side_effect = [
        15.0,  # outer start
        17.0,  # inner start
        17.7,  # inner finished (0.7s elapsed)
        18.2,  # outer finished (3.2s elapsed)
        Exception("Should not have been called again!"),
    ]

    swh_storage.content_add([StorageData.content, StorageData.content2])
    for call in timing.call_args_list:
        assert (
            call.args[0] != "swh_storage_masking_overhead_seconds"
        ), "Overhead metric sent unexpectedly"

    assert (
        swh_storage.content_get_data(cast(HashDict, StorageData.content.hashes()))
        is not None
    )

    timing.assert_any_call(
        "swh_storage_masking_overhead_seconds",
        # 3.2 s total - 0.7 s inner = 2500 ms overhead
        pytest.approx(2500.0),
        tags={"endpoint": "content_get_data"},
    )


def test_proxy_config_deprecation(masking_db_postgresql, swh_storage_backend):
    with pytest.warns(DeprecationWarning):
        assert MaskingProxyStorage(
            masking_db=masking_db_postgresql.info.dsn, storage=swh_storage_backend
        )
